/*

Copyright 1993, 1994, Cornell University

Cornell hereby grants permission to use, copy, modify, and distribute this program for any purpose 
and without fee, provided that these copyright and permission notices appear on all copies and 
supporting documentation, the name of Cornell not be used in advertising or publicity pertaining 
to distribution of the program without specific prior permission, notice be given in supporting 
documentation that copying and distribution is by permission of Cornell.  CORNELL MAKES NO 
REPRESENTATIONS OR WARRANTEES, EXPRESS OR IMPLIED.  By way of example, but not limitation, 
CORNELL MAKES NO REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR 
PURPOSE OR THAT THE USE OF THIS SOFTWARE OR DOCUMENTATION WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS, 
TRADEMARKS, OR OTHER RIGHTS.  Cornell shall not be held liable for any liability with respect to 
any claim by the user or any other party arising from use of the program.

This material is partially based on work sponsored by the National Science Foundation under Cooperative 
Agreement No. NCR-9318337.  The government has certain rights in this material.

*/



#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "reflect.h"
#include "refmon.h"

unsigned short clincnt = 1;

extern short debug;
extern unsigned long restrict_list[];
extern short restrict_cnt;

extern unsigned long deny_list[];
extern short deny_cnt;

extern unsigned long bcc_server_list[];
extern short bcc_srv_cnt;

extern unsigned long bcc_client_list[];
extern short bcc_clt_cnt;

extern unsigned long ref_ucast_list[];
extern short ref_ucast_cnt;          

extern short self_reflect,local_senders;
extern char *tracefile;

extern char conf_buf[];
extern char motd_buf[];
extern char mp_buf[];
extern char ci_buf[];
extern char ad_buf[];
extern char mv_mac_buf[];
extern char mv_pc_buf[];
extern char dy_buf[];
extern short motd_len;
extern int maven_port;
extern int log_limit;
extern int conference_id;
extern int maxallowed;
extern int min_mac_version;
extern int min_pc_version;
extern unsigned long control_ip;
extern unsigned long god_ip;

#ifdef MULTI
extern struct sockaddr_in nv_mcast;
extern struct sockaddr_in rfin_mcast;
extern struct sockaddr_in rfout_mcast;
extern struct sockaddr_in inout_mcast;
extern int nv_ttl,rf_ttl,inout_ttl;
#endif


FILE  *conf_fd;
short scan_str();

load_config(fname)
   char *fname;
{
   int            icnt;
   char           code[16],line[101],buf[100],*configfile = CONFIGFILE;
   int            cnt,hb,mh,ml,lb;
   char           *tmp;
   short          status;

   if (fname != NULL) 
      configfile = fname; 

   printf("\nVersion: %s\n",VERSION_NUM);

   if ((conf_fd = fopen (configfile, "r")) == NULL)
      return;

   tmp = conf_buf;

   while (!feof (conf_fd)) 
   {
      if ((code[0] = fgetc(conf_fd)) <= ' ') 
      {
         if (code[0] == '\n')
            ++clincnt;
         continue;
      }

      if ((icnt = fscanf (conf_fd, "%s", code + 1)) == EOF)
         break;

      if (icnt != 1)
      {
         printf("Unable to parse configuration code line %d\n",clincnt);
         exit(-1);
      }

      if (code[0] == ';')
      {
         fgets (line, 100, conf_fd);
         ++clincnt;
         continue;
      }

      printf("\n%d: ",clincnt);

      printf("%s ",code);


      if (strcasecmp (code, "DEBUG") == 0) 
      {
	 strcpy(tmp,"debug\n ");
	 tmp += strlen(tmp);
         debug = 1;
         continue;
      }

      if (strcasecmp(code, "MC-TO-NV") == 0) 
      {
#ifdef MULTI
	 strcpy(tmp,"mc-to-nv ");
	 tmp += strlen(tmp);
         if ((cnt = fscanf(conf_fd, "%i %i.%i.%i.%i",&nv_ttl,&hb,&mh,&ml,&lb)) != 5)
	 {
            printf ("MC-TO-NV paramater count error\n");
            exit(-1);
         }
         printf("%d %d.%d.%d.%d ",nv_ttl,hb,mh,ml,lb);
         sprintf(buf,"%d.%d.%d.%d ",hb,mh,ml,lb);

         sprintf(tmp,"%d.%d.%d.%d\n",hb,mh,ml,lb);
	 tmp += strlen(tmp);

         if ((nv_mcast.sin_addr.s_addr = inet_addr(buf)) == -1)
         {
            printf("inet_addr: bad MC-TO-NV multicast address\n");
            exit(-1);
         }
         nv_mcast.sin_family = AF_INET;
         nv_mcast.sin_port = htons(NV_PORT);
#else
         printf ("This copy of the reflector was not built with Multicast support\n");
         exit(-1);
#endif
         continue;
      }

      if (strcasecmp(code, "MC-OUT") == 0) 
      {
#ifdef MULTI
	 strcpy(tmp,"mc-out ");
	 tmp += strlen(tmp);

         if ((cnt = fscanf(conf_fd, "%i %i.%i.%i.%i",&rf_ttl,&hb,&mh,&ml,&lb)) != 5)
	 {
            printf ("MC-OUT-out paramater count error\n");
            exit(-1);
         }
         printf("%d %d.%d.%d.%d ",rf_ttl,hb,mh,ml,lb);
         sprintf(buf,"%d.%d.%d.%d ",hb,mh,ml,lb);

         sprintf(tmp,"%d.%d.%d.%d\n",hb,mh,ml,lb);
	 tmp += strlen(tmp);

         if ((rfout_mcast.sin_addr.s_addr = inet_addr(buf)) == -1)
         {
            printf("inet_addr: bad MC-OUT multicast address\n");
            exit(-1);
         }
         rfout_mcast.sin_family = AF_INET;
         rfout_mcast.sin_port = htons(RF_PORT);
#else
         printf ("This copy of the reflector was not built with Multicast support\n");
         exit(-1);
#endif
         continue;
      }

      if (strcasecmp(code, "MC-IN") == 0) 
      {
#ifdef MULTI
	 strcpy(tmp,"mc-in ");
	 tmp += strlen(tmp);

         if ((cnt = fscanf(conf_fd, "%i.%i.%i.%i",&hb,&mh,&ml,&lb)) != 4)
	 {
            printf ("MC-IN paramater count error\n", cnt);
            exit(-1);
         }
         printf("%d.%d.%d.%d ",hb,mh,ml,lb);
         sprintf(buf,"%d.%d.%d.%d ",hb,mh,ml,lb);

         sprintf(tmp,"%d.%d.%d.%d\n",hb,mh,ml,lb);
	 tmp += strlen(tmp);

         if ((rfin_mcast.sin_addr.s_addr = inet_addr(buf)) == -1)
         {
            printf("inet_addr: bad MC-IN multicast address\n");
            exit(-1);
         }
         rfin_mcast.sin_family = AF_INET;
         rfin_mcast.sin_port = htons(RF_PORT);
#else
         printf ("This copy of the reflector was not built with Multicast support\n");
         exit(-1);
#endif
         continue;
      }

      if (strcasecmp(code, "MC-GROUP") == 0) 
      {
#ifdef MULTI
	 strcpy(tmp,"mc-group ");
	 tmp += strlen(tmp);

         if ((cnt = fscanf(conf_fd, "%i %i.%i.%i.%i",&inout_ttl,&hb,&mh,&ml,&lb)) != 5)
	 {
            printf ("MC-GROUP paramater count error\n");
            exit(-1);
         }
         printf("%d %d.%d.%d.%d ",inout_ttl,hb,mh,ml,lb);
         sprintf(buf,"%d.%d.%d.%d ",hb,mh,ml,lb);

         sprintf(tmp,"%d.%d.%d.%d\n",hb,mh,ml,lb);
	 tmp += strlen(tmp);

         if ((inout_mcast.sin_addr.s_addr = inet_addr(buf)) == -1)
         {
            printf("inet_addr: bad MC-GROUP multicast address\n");
            exit(-1);
         }
         inout_mcast.sin_family = AF_INET;
         inout_mcast.sin_port = htons(RF_PORT);
#else
         printf ("This copy of the reflector was not built with Multicast support\n");
         exit(-1);
#endif
         continue;
      }

      if (strcasecmp(code, "MAX-PARTICIPANTS") == 0) 
      {
	 strcpy(tmp,"max-participants ");
	 tmp += strlen(tmp);

         if ((cnt = fscanf(conf_fd, "%i",&maxallowed)) != 1)
	 {
            printf ("MAX-PARTICIPANTS paramater count error\n");
            exit(-1);
         }
	 printf("%d ",maxallowed);
	 sprintf(buf,"%d ",maxallowed);

	 sprintf(tmp,"%d",maxallowed);
	 tmp += strlen(tmp);

	 if (maxallowed > MAXCLIENT)
	 {
	    printf("Maximum number of clients allowed is %d\n",MAXCLIENT);
	    maxallowed = MAXCLIENT;
         }

         status = scan_str(mp_buf,MSGBUF);

         if (status == -1)
         {
            printf ("max-participants message string does not end with // \n");
            exit(-1);
         }

         if (status == -2)
         {
            printf ("max-participants message length is greater then %d characters \n",MSGBUF);
            exit(-1);
         }

	 sprintf(tmp,"%s\n",mp_buf);
	 tmp += strlen(tmp);
         continue;

      }

      if (strcasecmp(code, "MIN-MAC-VERSION") == 0) 
      {
	 strcpy(tmp,"min-mac-version ");
	 tmp += strlen(tmp);

         if ((cnt = fscanf(conf_fd, "%i",&min_mac_version)) != 1)
	 {
            printf ("MIN-MAC-VERSION paramater count error\n");
            exit(-1);
         }
	 printf("%d ",min_mac_version);

	 sprintf(buf,"%d ",min_mac_version);

	 sprintf(tmp,"%d ",min_mac_version);
	 tmp += strlen(tmp);

         status = scan_str(mv_mac_buf,MSGBUF);

         if (status == -1)
         {
            printf ("min-mac-version message string does not end with // \n");
            exit(-1);
         }

         if (status == -2)
         {
            printf ("min-mac-version message length is greater then %d characters \n",MSGBUF);
            exit(-1);
         }

         sprintf(tmp,"%s\n",mv_mac_buf);
         tmp += strlen(tmp);

         continue;
      }

      if (strcasecmp(code, "MIN-PC-VERSION") == 0) 
      {
	 strcpy(tmp,"min-pc-version ");
	 tmp += strlen(tmp);

         if ((cnt = fscanf(conf_fd, "%i",&min_pc_version)) != 1)
	 {
            printf ("MIN-PC-VERSION paramater count error\n");
            exit(-1);
         }
	 printf("%d ",min_pc_version);

	 sprintf(buf,"%d ",min_pc_version);

	 sprintf(tmp,"%d ",min_pc_version);
	 tmp += strlen(tmp);

         status = scan_str(mv_pc_buf,MSGBUF);

         if (status == -1)
         {
            printf ("min-pc-version message string does not end with // \n");
            exit(-1);
         }

         if (status == -2)
         {
            printf ("min-pc-version message length is greater then %d characters \n",MSGBUF);
            exit(-1);
         }

         sprintf(tmp,"%s\n",mv_pc_buf);
         tmp += strlen(tmp);

         continue;
      }

      if (strcasecmp(code, "REFMON") == 0) 
      {
	 strcpy(tmp,"refmon ");
	 tmp += strlen(tmp);

         if ((cnt = fscanf(conf_fd, "%i.%i.%i.%i",&hb,&mh,&ml,&lb)) != 4)
	 {
            printf ("REFMON paramater count error\n");
            exit(-1);
         }
         printf("%d.%d.%d.%d ",hb,mh,ml,lb);
         sprintf(buf,"%d.%d.%d.%d ",hb,mh,ml,lb);
	 
         sprintf(tmp,"%d.%d.%d.%d\n",hb,mh,ml,lb);
	 tmp += strlen(tmp);

         if ((control_ip = inet_addr(buf)) == -1)
         {
            printf("inet_addr: bad REFMON address\n");
            exit(-1);
         }
         continue;
      }

      if (strcasecmp(code, "CONF-ID") == 0) 
      {
	 strcpy(tmp,"conf-id ");
	 tmp += strlen(tmp);

         if ((cnt = fscanf(conf_fd, "%i",&conference_id)) != 1)
	 {
            printf ("CONF-ID paramater count error\n");
            exit(-1);
         }
	 printf("%d ",conference_id);

	 sprintf(tmp,"%d ",conference_id);
	 tmp += strlen(tmp);

         status = scan_str(ci_buf,MSGBUF);

         if (status == -1)
         {
            printf ("conf-id message string does not end with // \n");
            exit(-1);
         }

         if (status == -2)
         {
            printf ("conf-id message length is greater then %d characters \n",MSGBUF);
            exit(-1);
         }

         sprintf(tmp,"%s\n",ci_buf);
         tmp += strlen(tmp);

         continue;

      }

      if (strcasecmp(code, "ADMIT") == 0) 
      {
	 strcpy(tmp,"admit ");
	 tmp += strlen(tmp);

         if (restrict_cnt == maxallowed)
         {
            printf("%d is the maximum number of of clients that can be added to the admit list\n",maxallowed);
            exit(-1);
         }

         if ((cnt = fscanf(conf_fd, "%i.%i.%i.%i",&hb,&mh,&ml,&lb)) != 4)
	 {
            printf ("ADMIT paramater count error\n");
            exit(-1);
         }
         printf("%d.%d.%d.%d ",hb,mh,ml,lb);
         sprintf(buf,"%d.%d.%d.%d ",hb,mh,ml,lb);
	 
         sprintf(tmp,"%d.%d.%d.%d ",hb,mh,ml,lb);
	 tmp += strlen(tmp);

         status = scan_str(ad_buf,MSGBUF);

         if (status == -1)
         {
            printf ("admit message string does not end with // \n");
            exit(-1);
         }

         if (status == -2)
         {
            printf ("admit message length is greater then %d characters \n",MSGBUF);
            exit(-1);
         }

         sprintf(tmp,"%s\n",ad_buf);
         tmp += strlen(tmp);


         if ((restrict_list[restrict_cnt++] = inet_addr(buf)) == -1)
         {
            printf("inet_addr: bad ADMIT address\n");
            exit(-1);
         }
         continue;
      }

      if (strcasecmp(code, "DENY") == 0) 
      {
	 strcpy(tmp,"deny ");
	 tmp += strlen(tmp);

         if (deny_cnt == maxallowed)
         {
            printf("%d is the maximum number of of clients that can be added to the deny list\n",maxallowed);
            exit(-1);
         }

         if ((cnt = fscanf(conf_fd, "%i.%i.%i.%i",&hb,&mh,&ml,&lb)) != 4)
	 {
            printf ("DENY paramater count error\n");
            exit(-1);
         }
         printf("%d.%d.%d.%d ",hb,mh,ml,lb);
         sprintf(buf,"%d.%d.%d.%d ",hb,mh,ml,lb);
	 
         sprintf(tmp,"%d.%d.%d.%d ",hb,mh,ml,lb);
	 tmp += strlen(tmp);

         status = scan_str(dy_buf,MSGBUF);

         if (status == -1)
         {
            printf ("deny message string does not end with // \n");
            exit(-1);
         }

         if (status == -2)
         {
            printf ("deny message length is greater then %d characters \n",MSGBUF);
            exit(-1);
         }

         sprintf(tmp,"%s\n",dy_buf);
         tmp += strlen(tmp);

         if ((deny_list[deny_cnt++] = inet_addr(buf)) == -1)
         {
            printf("inet_addr: bad DENY address\n");
            exit(-1);
         }
         continue;
      }


      if (strcasecmp(code, "SELF-REFLECT") == 0) 
      {
	 strcpy(tmp,"self-reflect\n ");
	 tmp += strlen(tmp);
         self_reflect = 1;
         continue;
      }

      if (strcasecmp(code, "NO-LOCAL-SENDERS") == 0) 
      {
	 strcpy(tmp,"no-local-senders\n");
	 tmp += strlen(tmp);
         local_senders = 0;
         continue;
      }

      if (strcasecmp(code, "LOG") == 0) 
      {
	 strcpy(tmp,"log ");
	 tmp += strlen(tmp);

         if ((cnt = fscanf(conf_fd, "%s",tracefile)) != 1)
	 {
            printf ("LOG paramater count error\n");
            exit(-1);
         }
	 printf("%s ",tracefile);

	 sprintf(tmp,"%s\n",tracefile);
	 tmp += strlen(tmp);

         continue;
      }

      if (strcasecmp(code, "OBTAIN-BCC") == 0) 
      {
	 strcpy(tmp,"obtain-bcc ");
	 tmp += strlen(tmp);

         if (bcc_srv_cnt == MAXSRCREF)
         {
            printf("%d is the maximum number of OBTAIN-BCC reflectors allowed\n",MAXSRCREF);
            exit(-1);
         }

         if ((cnt = fscanf(conf_fd, "%i.%i.%i.%i",&hb,&mh,&ml,&lb)) != 4)
	 {
            printf ("OBTAIN-BCC paramater count error\n");
            exit(-1);
         }

         printf("%d.%d.%d.%d ",hb,mh,ml,lb);
         sprintf(buf,"%d.%d.%d.%d ",hb,mh,ml,lb);
	 
         sprintf(tmp,"%d.%d.%d.%d\n",hb,mh,ml,lb);
	 tmp += strlen(tmp);

         if ((bcc_server_list[bcc_srv_cnt++] = inet_addr(buf)) == -1)
         {
            printf("inet_addr: bad OBTAIN-BCC address\n");
            exit(-1);
         }

         continue;
      }

      if (strcasecmp(code, "ADMIT-BCC-CLIENT") == 0) 
      {
	 strcpy(tmp,"admit-bcc-client ");
	 tmp += strlen(tmp);

         if (bcc_clt_cnt == MAXSRCREF)
         {
            printf("%d is the maximum number of ADMIT-BCC-CLIENT reflectors allowed\n",MAXSRCREF);
            exit(-1);
         }

         if ((cnt = fscanf(conf_fd, "%i.%i.%i.%i",&hb,&mh,&ml,&lb)) != 4)
	 {
            printf ("ADMIT-BCC-CLIENT paramater count error\n");
            exit(-1);
         }

         printf("%d.%d.%d.%d ",hb,mh,ml,lb);
         sprintf(buf,"%d.%d.%d.%d ",hb,mh,ml,lb);

         sprintf(tmp,"%d.%d.%d.%d\n",hb,mh,ml,lb);
	 tmp += strlen(tmp);

         if ((bcc_client_list[bcc_clt_cnt++] = inet_addr(buf)) == -1)
         {
            printf("inet_addr: bad ADMIT-BCC-CLIENT address\n");
            exit(-1);
         }
         continue;
      }

      if (strcasecmp(code, "UNICAST-REF") == 0) 
      {
	 strcpy(tmp,"unicast-ref ");
	 tmp += strlen(tmp);

         if (ref_ucast_cnt == MAXSRCREF)
         {
            printf("%d is the maximum number of UNICAST-REF reflectors allowed\n",MAXSRCREF);
            exit(-1);
         }

         if ((cnt = fscanf(conf_fd, "%i.%i.%i.%i",&hb,&mh,&ml,&lb)) != 4)
	 {
            printf ("UNICAST-REF paramater count error\n");
            exit(-1);
         }

         printf("%d.%d.%d.%d ",hb,mh,ml,lb);
         sprintf(buf,"%d.%d.%d.%d ",hb,mh,ml,lb);
	 
         sprintf(tmp,"%d.%d.%d.%d\n",hb,mh,ml,lb);
	 tmp += strlen(tmp);

         if ((ref_ucast_list[ref_ucast_cnt++] = inet_addr(buf)) == -1)
         {
            printf("inet_addr: bad UNICAST-REF address\n");
            exit(-1);
         }

         continue;
      }

      if (strcasecmp(code, "MAVEN") == 0) 
      {
	 strcpy(tmp,"maven ");
	 tmp += strlen(tmp);

         if ((cnt = fscanf(conf_fd, " %i", &maven_port)) != 1)
	 {
            printf ("MAVEN paramater count error\n");
            exit(-1);
         }

         printf("%d ",maven_port);
         sprintf(buf,"%d ",maven_port);

         sprintf(tmp,"%d\n",maven_port);
	 tmp += strlen(tmp);

         continue;
      }

      if (strcasecmp(code, "LOG-LIMIT") == 0) 
      {
	 strcpy(tmp,"log-limit ");
	 tmp += strlen(tmp);

         if ((cnt = fscanf(conf_fd, " %i", &log_limit)) != 1)
	 {
            printf ("LOG-LIMIT paramater count error\n");
            exit(-1);
         }

         printf("%d ",log_limit);
         sprintf(buf,"%d ",log_limit);

         sprintf(tmp,"%d\n",log_limit);
	 tmp += strlen(tmp);

         continue;
      }

      if (strcasecmp(code, "CONF-MGR") == 0) 
      {
         strcpy(tmp, "conf-mgr ");
         tmp += strlen(tmp);

         if ((cnt = fscanf(conf_fd, "%i.%i.%i.%i", &hb,&mh,&ml,&lb)) != 4)
	 {
            printf("CONF-MGR parameter count error\n");
            exit(-1);
          }

          printf("%d.%d.%d.%d ",hb,mh,ml,lb);
          sprintf(buf, "%d.%d.%d.%d ",hb,mh,ml,lb);

          sprintf(tmp, "%d.%d.%d.%d\n",hb,mh,ml,lb);
          tmp += strlen(tmp);

          if ((god_ip = inet_addr(buf)) == -1)
	  {
            printf("inet_addr: bad CONF-MGR address\n");
            exit(-1);
          }

          continue;
       }




      if (strcasecmp(code, "MOTD") == 0) 
      {
	 strcpy(tmp,"motd \n");
	 tmp += strlen(tmp);

         status = scan_str(motd_buf,MOTDBUF);

         if (status == -1)
         {
            printf ("MOTD message does not end with // \n");
            exit(-1);
         }

         if (status == -2)
         {
            printf ("maximum MOTD message length s %d\n",MOTDBUF);
            exit(-1);
         }

         sprintf(tmp,"%s\n",motd_buf);
         motd_len = strlen(motd_buf);
	 tmp += motd_len;

	 continue;
      }

      printf ("Configuration file is damaged: Bad CODE value line %d\n",clincnt);
      exit(-1);
   }
   printf("\n");
   fflush(stdout);
   close(conf_fd);
}


short scan_str(buf,maxlen)
   char *buf;
   short maxlen;
{
   short          len;
   char           c,c1,c2;

   len = 0;

   while (1)
   {
      if ((c = fgetc(conf_fd)) == EOF)
         return(-1);

      if ((c == '\n') || (c == '\r'))
      {
         if ((c1 = fgetc(conf_fd)) == EOF)
            return(-1);

         if (c1 == '/')
         {
            if ((c2 = fgetc(conf_fd)) == EOF)
               return(-1);

            if (c2 == '/')
               return(0);

            if ((len = len+3) >= maxlen)
               return(-2);

            *buf++ = c;
            printf("%c",c);
   
            *buf++ = c1;
            printf("%c",c1);

            *buf++ = c2;
            printf("%c",c2);
	 }
	 else
	 {
            if ((len = len+2) >= maxlen)
               return(-2); 

            *buf++ = c;
            printf("%c",c);

            *buf++ = c1;
            printf("%c",c1);
	 }
      }
      else
      {
         if ((len = len+1) >= maxlen)
            return(-2);

         *buf++ = c;
         printf("%c",c);
      }
   }
}
