/*

Copyright 1993, 1994, Cornell University

Cornell hereby grants permission to use, copy, modify, and distribute this program for any purpose 
and without fee, provided that these copyright and permission notices appear on all copies and 
supporting documentation, the name of Cornell not be used in advertising or publicity pertaining 
to distribution of the program without specific prior permission, notice be given in supporting 
documentation that copying and distribution is by permission of Cornell.  CORNELL MAKES NO 
REPRESENTATIONS OR WARRANTEES, EXPRESS OR IMPLIED.  By way of example, but not limitation, 
CORNELL MAKES NO REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR 
PURPOSE OR THAT THE USE OF THIS SOFTWARE OR DOCUMENTATION WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS, 
TRADEMARKS, OR OTHER RIGHTS.  Cornell shall not be held liable for any liability with respect to 
any claim by the user or any other party arising from use of the program.

This material is partially based on work sponsored by the National Science Foundation under Cooperative 
Agreement No. NCR-9318337.  The government has certain rights in this material.

*/

#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/time.h>
#include <netinet/in.h>

#include "reflect.h"
#include "refmon.h"

void                proc_sig();
void                do_timer();
void                my_perror();
client              *find_client(); 
maven_client        *find_maven_client(); 
maven_client        *new_maven_client();
client              *open_connection(); 
short               get_type();

client *chead = NULL;                         /* ptr to the active client list             */
client *clfree = NULL;                        /* ptr to the free client list               */

slist *sfree = NULL;                          /* ptr to the free slist list                */

short cfree_cnt,sfree_cnt;                    /* # of free clients and slists              */

maven_client *mhead = NULL;                   /* global ptr to the maven client list       */

unsigned long restrict_list[MAXCLIENT];       /* list of IP addrs if this reflector        */
short restrict_cnt = 0;                       /* is restricted                             */

unsigned long deny_list[MAXCLIENT];           /* list of IP addrs to deny connectivity to  */
short deny_cnt = 0;                           /*                                           */

unsigned long bcc_server_list[MAXSRCREF];     /* list of Blind Carbon Copy Servers, from   */
short bcc_srv_cnt = 0;                        /* whom we are to get video from             */

unsigned long bcc_client_list[MAXSRCREF];     /* list of Blind Carbon Copy Clients, to     */
short bcc_clt_cnt = 0;                        /* whom we are to give video to              */

unsigned long ref_ucast_list[MAXSRCREF];      /* list of unicast reflectors                */
short ref_ucast_cnt = 0;                      /* count of unicast reflectors               */

char *tracefile;                              /* trace / log file name                     */
FILE *log_file;

short debug = 0;                              /* if 1, turn on debugging messages          */
short self_reflect = 0;                       /* if 1, reflect back to the send            */

short local_senders = 1;                      /* if 0, disallow local conferencing, used   */
					      /* if this reflector is simply feeding a     */
                                              /* conference to all of it's clients         */

unsigned long control_ip = -1;                /* ip address for refmon                     */
unsigned long god_ip = -1;                    /* ip address for the conference god         */


int conference_id = 0;                        /* conference id                             */

short timer_expired = 0;                      /* keep track of our timer                   */

int min_mac_version = 0;                      /* minimum mac version # required for this ref  */ 
int min_pc_version = 0;                       /* minimum pc version # required for this ref   */ 

int vid_sock,cntrl_sock,msg_sock;             /* video socket, control                     */
				              /* socket, and message socket for supporing  */
					      /* the control socket                        */

struct sockaddr_in myaddr;                    /* my address                                */
int maxallowed;                               /* maximum # of clients allowed              */

char conf_buf[CONFBUF];	                      /* buffer for storing the config file        */  
char motd_buf[MOTDBUF];                       /* buffer for storing the MOTD               */

char mp_buf[MSGBUF];                          /* buffer for max-participants error msg     */
char ci_buf[MSGBUF];                          /* buffer for conference-id error msg        */
char ad_buf[MSGBUF];                          /* buffer for admit error msg                */
char mv_mac_buf[MSGBUF];                      /* buffer for min-mac-version error msg      */
char mv_pc_buf[MSGBUF];                       /* buffer for min-pc-version error msg       */
char dy_buf[MSGBUF];                          /* buffer for deny error msg                 */

short motd_len = 0;

int maven_port = 0;                           /* sockets for handling Maven clients        */
int maven_sock = 0;
int maven_cntl_sock = 0;

unsigned short client_cnt = 0;                /* # of Mac/PC clients                       */

unsigned short bcc_client_cnt = 0;            /* # of BCC clients                          */
unsigned short bcc_origin_cnt = 0;            /* # of BCC origins                          */
unsigned short bcc_server_cnt = 0;            /* # of BCC servers                          */

unsigned short ref1_server_cnt = 0;           /* # of REF1 servers                         */
unsigned short ref1_origin_cnt = 0;           /* # of REF1 origins                         */
unsigned short ref1_client_cnt = 0;           /* # of REF1 clients                         */

unsigned short ref2_server_cnt = 0;           /* # of REF2 servers                         */
unsigned short ref2_origin_cnt = 0;           /* # of REF2 origins                         */

unsigned short ref3_server_cnt = 0;           /* # of REF3 servers                         */
unsigned short ref3_origin_cnt = 0;           /* # of REF3 origins                         */


int log_limit = 10000;                        /* default log file line limit               */

char start_time[25];                          /* date and time when reflector was started  */

#ifdef MULTI
struct sockaddr_in nv_mcast;                  /* socket for sending to nv                  */
int nv_mcast_sock;

struct sockaddr_in rfout_mcast;               /* socket for sending to other reflectors    */
int rfout_mcast_sock;

struct sockaddr_in rfin_mcast;                /* socket for receiving from other reflectors */ 
int rfin_mcast_sock;

struct sockaddr_in inout_mcast;               /* socket for sending and receiving from other reflectors */ 
int inout_mcast_sock;

int nv_ttl,rf_ttl,inout_ttl;                  /* multicast ttl's                           */
#endif

unsigned char  buffer[MAXMSG];

main(argc, argv)
    int                argc;
    char               *argv[];
{

    unsigned char       *msg,*s,*cptr,*cptr1,version;
    VideoPacketHeader   *vidptr, *vtmp; 
    RefConPkt           pkt;
    RefConPkt           *conpkt;
    client              *cltptr,*ctmp;
    maven_client        *mcltptr,*mtmp;
    struct sockaddr_in  csock,clnt_addr;
    struct in_addr      in;
    int                 msglen;
    short 		type; 
    struct timeval      tp;
    struct timezone     tzp;
    char                *tmp;


    argc--; argv++;

    tracefile = LOGFILE;
    maxallowed = MAXCLIENT;

    if (argc > 0)
       load_config(*argv);
    else
       load_config(NULL);

    if (log_limit != 0)
    {

       if (strcmp(tracefile,"stdout") == 0)
	  log_file = stdout;
       else
       {
          if ((log_file = fopen(tracefile,"w")) != NULL)
          {
             setlinebuf(log_file);
             dolog("open_log file: %s\n",tracefile);
          }
          else
             log_limit = 0;
       }
    }

    if (strlen(ci_buf) == 0)
       strcpy("Sorry, the Conference ID has changed.  Please reconnect with the correct one.",ci_buf);

    if (strlen(mp_buf) == 0)
       strcpy("Sorry, maximum number of users exceeded.",mp_buf);

    signal(SIGTERM, proc_sig);  
    signal(SIGINT, proc_sig);    

    get_my_addr(&myaddr);
    if (myaddr.sin_addr.s_addr == 0)
    {
       dolog("Unable to get a internet address\n");
       exit(1);
    }

    init_mem();
    init_socket();
    init_timer();
    open_bcc_servers();

    gettimeofday(&tp, &tzp);
    tmp = ctime(&tp.tv_sec);
    tmp[24] = ' ';
    bcopy(tmp,start_time,strlen(tmp));

    msg = &buffer[40];
    vidptr = (VideoPacketHeader *) msg;
    while (1)
    {
       type = receive(msg,&msglen,&csock);

       if (deny(csock))
       {
          clnt_addr.sin_family = AF_INET;
	  clnt_addr.sin_port = htons(VID_PORT);
	  bcopy(&vidptr->routing.src.addr,&clnt_addr.sin_addr,4);
          write_msg(&clnt_addr,kMessageType1,dy_buf); 
	  continue;
       }

       switch (type)
       {

	  case MAVEN_CNTL:
	      if ((mcltptr = find_maven_client(csock.sin_addr.s_addr)) == NULL)
	      {
	         if ((mcltptr = new_maven_client(csock.sin_addr.s_addr)) == NULL)
		    break;
                 dolog("New Maven client at %s\n", inet_ntoa(csock.sin_addr));
              }

              mcltptr->mvn_rtimer = 0;

	      vat_id(mcltptr);

	      break;

	  case MAVEN:
	      if ((mcltptr = find_maven_client(csock.sin_addr.s_addr)) == NULL)
		 break;

              mcltptr->mvn_rtimer = 0;
              mcltptr->mvn_seq++;

	      if (mcltptr->mvn_talker == 0)
	         dolog("maven at %s is speaking\n",inet_ntoa(csock.sin_addr));

              if (mcltptr->mvn_talker++ > 50)
                 mcltptr->mvn_talker = 0;

              mtmp = mhead;
              while (mtmp != NULL)
              {
                 if ((self_reflect == 1) || (mcltptr != mtmp))
                    if (sendto(maven_sock,msg,msglen,0,&mtmp->mvn_addr,sizeof(struct sockaddr_in)) != msglen)
		       dolog("Maven sendto error\n");
                 mtmp = mtmp->mvn_nptr;
              }

	      /* 
		 do not send Maven packets to CU-SeeMe clients, unless that Maven client is also 
		 a CU-SeeMe client.
              */

	      if ((cltptr = find_client(csock.sin_addr.s_addr)) == NULL)
		 break;

              vtmp = (VideoPacketHeader *) (((unsigned char *) msg) - HEADERLEN);

              vtmp->routing.dest.family = htons(kGroup);           
              vtmp->routing.dest.port = htons(VID_PORT);           
              vtmp->routing.src.family = htons(kGroup);          
              vtmp->routing.src.port  = htons(VID_PORT);          

	      cptr = (unsigned char *) &vtmp->routing.src.addr;
	      cptr1 = (unsigned char *) &csock.sin_addr;
	      *cptr++ = *cptr1++;
	      *cptr++ = *cptr1++;
	      *cptr++ = *cptr1++;
	      *cptr++ = *cptr1++;

	      cptr = (unsigned char *) &vtmp->seqNum;
              cptr1 = (unsigned char *) &mcltptr->mvn_seq;
	      *cptr++ = *cptr1++;
	      *cptr++ = *cptr1++;
	      *cptr++ = *cptr1++;
	      *cptr++ = *cptr1++;

	      vtmp->message = htons(kAudio);
              vtmp->dataType = htons(kAudio);
	      vtmp->len = htons(msglen + HEADERLEN);

              ctmp = chead;
              while (ctmp != NULL)
              {
	         if ((ctmp->clnt_config.flags & REC_AUDIO) && (ctmp->clnt_config.flags & WANT_LURCKERS))
                    if (sendto(vid_sock,(char *)vtmp,msglen+HEADERLEN,0,&ctmp->clnt_addr,sizeof(struct sockaddr_in)) 
			!= msglen+HEADERLEN)
		       dolog("Maven send to error in reflect\n");
                 ctmp = ctmp->clnt_nptr;
              }

             break;

	  case REF1VIDEO:
	  case REF2VIDEO:
          case VIDEO:

#ifdef DEBUG
             if (debug)
             {
	        if (ntohs(vidptr->message) == kAudio)
	  	   printf("AUDIO\n");
                printf("video packet len %d network src %s ", msglen, inet_ntoa(csock.sin_addr));
		in.s_addr = vidptr->routing.src.addr;
                printf("content src %s ",inet_ntoa(in));
                printf("family %d  seq %ld msg %d dtype %d\n", ntohs(vidptr->routing.dest.family),
			ntohl(vidptr->seqNum), ntohs(vidptr->message),ntohs(vidptr->dataType));
             }
#endif
	     /* JAL 5/11 move this check to the very beginning */

             if (vidptr->routing.src.addr == 0)
	     {
		dolog("received a packet with a 0 source address from %s\n",inet_ntoa(csock.sin_addr));
		break;
             }

	     if ((cltptr = find_client(csock.sin_addr.s_addr)) == NULL)
	     {	
#ifdef DEBUG
	        if (debug)
	           printf("client not found \n");
#endif
		if (type == VIDEO)
		{
		   if ((type = get_type(csock)) == -1)
		   {
		      dolog("client not found and packet being dropped due to address restrictions\n");
		      break;
                   }
                }
		else
		   if (type == REF1VIDEO)
		      type = REF1_SERVER;
                   else
		      type = REF2_SERVER;


                if (ntohs(vidptr->dataType) == kConfigRefType)
		{
		   if ((type != BCC_SERVER) && (type != BCC_CLIENT) && (type != REF3_SERVER))
		   {
		      dolog("Reflector configuration mis-match from %s\n",inet_ntoa(csock.sin_addr));
		      break;

                   }
		}
		else
                   if ((ntohs(vidptr->dataType) != kConfigVideoType) || 
		       (ntohs(vidptr->message) != kOpenConnection))
                   {
		      dolog("client not found and initial message is not Open\n");
		      break;
		   }


                /*   added by MAG 6/27/94   */
                if ((csock.sin_addr.s_addr) == god_ip) 
                {
                   if (debug)
                      printf("This is the God-ip. Changing conf-id if necessary.\n");
                 
                   if (conference_id != ntohs(vidptr->conferenceid)) 
		   {
                      dolog("Conf-id was %d\n", conference_id);
 
                      conference_id = ntohs(vidptr->conferenceid);
 
                      dolog("Conf-id changed to %d\n", conference_id);
 
                      /* weed out all the clients who now have bad conf-id's          */
                      /* however if the god-ip changed it to 0, don't kick anybody off*/
 
                      if (conference_id != 0)
                         remove_some_clients(conference_id);
                   }
                }

		if ((type == CLIENT) && (conference_id != 0))
		   if (ntohs(vidptr->conferenceid) != conference_id)
		   {
		      dolog("conference ids do not match %d %d\n",ntohs(vidptr->conferenceid),conference_id);
		      clnt_addr.sin_family = AF_INET;
		      clnt_addr.sin_port = htons(VID_PORT);
		      bcopy(&vidptr->routing.src.addr,&clnt_addr.sin_addr,4);
                      write_msg(&clnt_addr,kMessageType1,ci_buf);
		      break;
                   }

                cptr = (unsigned char *) ((unsigned char *) vidptr + HEADERLEN) + 
		       sizeof(cltptr->clnt_config.clientCount) +
		       sizeof(cltptr->clnt_config.seqNum) +
		       sizeof(cltptr->clnt_config.name) +
		       sizeof(cltptr->clnt_config.sendMode) +
		       sizeof(cltptr->clnt_config.recvMode) +
		       sizeof(cltptr->clnt_config.flags);

	        version = *cptr;

                cptr1 = (unsigned char *) ((unsigned char *) vidptr + HEADERLEN) + 
		        sizeof(cltptr->clnt_config.clientCount) +
		        sizeof(cltptr->clnt_config.seqNum) +
		        sizeof(cltptr->clnt_config.name) +
		        sizeof(cltptr->clnt_config.sendMode) +
		        sizeof(cltptr->clnt_config.recvMode);

		/* assume version == 1 is a PC client     */

                if (((*cptr1 & PC_CLIENT) || (version == 1)) &&
		    (min_pc_version != 0))
	        {
		   if (version < min_pc_version)
		   {
		      dolog("old PC version # %d is being rejected\n",version);

		      clnt_addr.sin_family = AF_INET;
		      clnt_addr.sin_port = htons(VID_PORT);
		      bcopy(&vidptr->routing.src.addr,&clnt_addr.sin_addr,4);
                      write_msg(&clnt_addr,kMessageType1,mv_pc_buf);
		      break;
                   }
	        }
	        else
	           if (min_mac_version != 0) 
		   {
		      if (version < min_mac_version)
		      {
                         dolog("old MAC version # %d is being rejected\n",version);
                         clnt_addr.sin_family = AF_INET;
                         clnt_addr.sin_port = htons(VID_PORT);
                         bcopy(&vidptr->routing.src.addr,&clnt_addr.sin_addr,4);
                         write_msg(&clnt_addr,kMessageType1,mv_mac_buf);
                         break;
		      }	
		   }

                if ((cltptr = open_connection(vidptr,&csock,type)) != NULL)
		   distribute(vidptr,cltptr,TRUE);

		break;
	     }

#ifdef DEBUG
             if (debug)
             {
                if (cltptr->clnt_flags & CLIENT)        
                   printf("CLIENT\n");

                if (cltptr->clnt_flags & BCC_CLIENT)
                   printf("BCC_CLIENT\n");
                if (cltptr->clnt_flags & BCC_SERVER)  
                   printf("BCC_SERVER\n");
                if (cltptr->clnt_flags & BCC_ORIGIN)  
                   printf("BCC_ORIGIN\n");
               
                if (cltptr->clnt_flags & REF1_CLIENT)   
                   printf("REF1_CLIENT\n");
                if (cltptr->clnt_flags & REF1_SERVER)   
                   printf("REF1_SERVER\n");
                if (cltptr->clnt_flags & REF1_ORIGIN)   
                   printf("REF1_ORIGIN\n");
               
                if (cltptr->clnt_flags & REF2_SERVER)  
                   printf("REF2_SERVER\n");
                if (cltptr->clnt_flags & REF2_ORIGIN)  
                   printf("REF2_ORIGIN\n");

                if (cltptr->clnt_flags & REF3_SERVER)  
                   printf("REF3_SERVER\n");
                if (cltptr->clnt_flags & REF3_ORIGIN)  
                   printf("REF3_ORIGIN\n");
             }
#endif
               
#ifdef DEBUG
             if (debug)
	        printf("client found \n");
#endif
             cltptr->clnt_rtimer = 0;
             cltptr->clnt_flags &= ~SUSPEND;

	     /* if this is just a keep alive from a REF3 SERVER just return now */ 
	     if ((cltptr->clnt_flags & REF3_SERVER) && ((csock.sin_addr.s_addr == vidptr->routing.src.addr)))
		break;

	     /* if this is just a keep alive from a BCC SERVER just return now */ 
	     if ((cltptr->clnt_flags & BCC_SERVER) && ((csock.sin_addr.s_addr == vidptr->routing.src.addr)))
		break;

              
	      switch (ntohs(vidptr->routing.dest.family))
              {
                 case kReflector:

		    if ((ntohs(vidptr->dataType) == kConfigVideoType) && 
			(ntohs(vidptr->message) == kOpenConnection))
		    {
                       continue_connection(cltptr,vidptr);  
		       distribute(vidptr,cltptr,TRUE);
		    }
                    else
		       if ((ntohs(vidptr->dataType) == kConfigVideoType) && 
			   (ntohs(vidptr->message) == kCloseConnection))
		       {
			  if (cltptr->clnt_flags & (BCC_SERVER | REF1_SERVER | REF2_SERVER | REF3_SERVER))
			  {
		             if ((cltptr = find_client(vidptr->routing.src.addr)) == NULL)
			     {
		                in.s_addr = vidptr->routing.src.addr;
		                dolog("Unable to find server's client at %s for close\n",inet_ntoa(in));
				break;
			     }
			  }

                          dolog("closing connection from %s\n",cltptr->clnt_config.name);

		          distribute(vidptr,cltptr,TRUE);
                          delete_client(cltptr);
                       }
                    break;

                 case kClient:
		    if ((cltptr = find_client(vidptr->routing.dest.addr)) == NULL)
		    {
		       in.s_addr = vidptr->routing.dest.addr;
		       dolog("Unable to find the destination client at %s for a kClient mesg\n",inet_ntoa(in));
		       dolog("msg from %s\n",inet_ntoa(csock.sin_addr));
		       break;
		    }

                    write_pkt(vidptr,cltptr);
                    break;

                 case kGroup:
	            if (cltptr->clnt_flags & (BCC_SERVER | REF1_SERVER | REF2_SERVER | REF3_SERVER))
		    {
		       if ((cltptr = find_client(vidptr->routing.src.addr)) == NULL)
		       {
		          in.s_addr = vidptr->routing.src.addr;
		          dolog("Unable to find server's client at %s for group\n",inet_ntoa(in));
		          break;
		       }
	            }

	            if (ntohs(vidptr->message) == kAudio)
	               distribute_audio(vidptr,cltptr);
		    else
	               distribute(vidptr,cltptr,FALSE);

                    break;

		 default:
                    dolog("unknown video message type\n");
                    break;
              }
	      break;


           case CONTROL:
#ifdef DEBUG
              if (debug)
                 printf("process control packet\n");
#endif
              conpkt = (RefConPkt *) msg;

              if (msglen < MINREFPKT)
              {
                 dolog("msglen is less then MINREFPKT\n");
                 break;    
              }

              if (msglen < ntohs(conpkt->msg_len))
              {
                 dolog("conpkt->msg_len is less then msglen\n");
                 break;    
              }

              process_control_pkt(conpkt);

              break;

           default:
              dolog("bad value returned from receive\n");
              exit(1);
       }

       if (timer_expired) 
       {
          timer_expired--;
          do_timer();
       }
    }
}



init_mem()
{
     unsigned char *stash;
     client *cltptr;
     slist *sptr;
     int cnt;



     if ((stash = (unsigned char *) calloc(MAXCLIENT,sizeof(client))) == NULL)
     {
        dolog("Unable to get client memory\n");
        exit(1);
     }

     for (cnt = 0, cltptr = (client *) stash; cnt < MAXCLIENT; cnt++, cltptr++)
	free_client(cltptr);

     if ((stash = (unsigned char *) calloc((MAXCLIENT * MAXCLIENT * 2),sizeof(slist))) == NULL)
     {
        dolog("Unable to get slist memory \n");
        exit(1);
     }

     for (cnt = 0, sptr = (slist *) stash; cnt < (MAXCLIENT * MAXCLIENT *2); cnt++, sptr++)
	free_slist(sptr);
}




free_client(cltptr)
    client *cltptr;
{
    cltptr->clnt_nptr = clfree;
    clfree = cltptr;
    cfree_cnt++;
}


client *get_client()
{
   client *cltptr;

   if (cfree_cnt == 0)
   {
      dolog("No more free clients\n");
      exit(1);
   }

   if (clfree == NULL) 
   {
      dolog("Unexpectedly no more free clients\n");
      exit(1);
   }

   cltptr = clfree;
   clfree = clfree->clnt_nptr;
   cfree_cnt--;
   bzero(cltptr,sizeof(client));
   return(cltptr);
}

free_slist(sptr)
    slist *sptr;
{
    sptr->snd_nptr = sfree;
    sfree = sptr;
    sfree_cnt++;
}


slist *get_slist()
{
   slist *sptr;

   if (sfree_cnt == 0)
   {
      dolog("No more free slists\n");
      exit(1);
   }

   if (sfree == NULL) 
   {
      dolog("Unexpectedly no more free slists\n");
      exit(1);
   }

   sptr = sfree;
   sfree = sfree->snd_nptr;
   sfree_cnt--;
   bzero(sptr,sizeof(slist));
   return(sptr);
}


remove_some_clients(conf_id)
   int conf_id;
{
   client *cptr,*cptr1;

   cptr = chead;

   while (cptr != NULL) 
   {
      cptr1 = cptr->clnt_nptr;
      if ((cptr->clnt_flags & CLIENT) && (cptr->clnt_conf_id != conf_id)) 
      {
	 dolog("client %s is being deleted due to conf-id mis-match\n",cptr->clnt_config.name);
         write_msg(&cptr->clnt_addr,kMessageType1,ci_buf);
         delete_client(cptr);
      }
     cptr = cptr1;
   }
}
