/*

Copyright 1993, 1994, Cornell University

Cornell hereby grants permission to use, copy, modify, and distribute this program for any purpose 
and without fee, provided that these copyright and permission notices appear on all copies and 
supporting documentation, the name of Cornell not be used in advertising or publicity pertaining 
to distribution of the program without specific prior permission, notice be given in supporting 
documentation that copying and distribution is by permission of Cornell.  CORNELL MAKES NO 
REPRESENTATIONS OR WARRANTEES, EXPRESS OR IMPLIED.  By way of example, but not limitation, 
CORNELL MAKES NO REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR 
PURPOSE OR THAT THE USE OF THIS SOFTWARE OR DOCUMENTATION WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS, 
TRADEMARKS, OR OTHER RIGHTS.  Cornell shall not be held liable for any liability with respect to 
any claim by the user or any other party arising from use of the program.

This material is partially based on work sponsored by the National Science Foundation under Cooperative 
Agreement No. NCR-9318337.  The government has certain rights in this material.

*/



/* address families */

#define kGroup				0
#define kClient				1
#define kReflector			2

/* data types */
#define kHalfSizeVideoType		1
#define kFullSizeVideoType		2
#define kMaxVideoDataType		2
#define kAudio				3

#define kControlType			100
#define kConfigVideoType		101
#define kPacketLossReport		102
#define kAckType			103
#define kMessageType1                   104
#define kMessageType2                   105
#define kConfigRefType		        106

/* video control messages */
#define kNoMessage			0
	
#define kOpenConnection			1
#define kStartVideo			2
#define kModifyVideo			3
#define kRefuseVideo			4
#define kStopVideo			5
#define kCloseConnection		6

#define kGotOpen		-kOpenConnection
#define kGotStart		-kStartVideo
#define kGotModify		-kModifyVideo
#define kGotRefuse		-kRefuseVideo
#define kGotStop		-kStopVideo
#define kGotClose		-kCloseConnection

#define kContinueConnection		12
#define kSuspendMessage			13

#define kFrameEndMessage		20


/* connect mode */
#define kInternal			-1
#define kNoConnection			0
#define kWaitingForOpenAck		1
#define kWaitingForClientAck		2
#define kSuspended			3
#define kConnected			4

/* send/receive mode */
#define kNoVideoSource			-1
#define kNoSize				0
#define kHalfSize			1
#define kFullSize			2
#define kDefaultSize			100

/* copy mode */
#define kAsIs				0
#define kHalf				1
#define kDouble				2

/* grab frame mode */
#define kNoFrames			0
#define kGrab1Frame			1
#define kGrab2Frames			2

#define VERSION_NUM	               "2.30B1"
#define LOGFILE                         "reflect.log"
#define CONFIGFILE			"reflect.conf"
#define DENY_ACCESS			"You are not permitted access to this reflector"

#define VID_PORT 			7648
#define CONTROL_PORT 			7649   
#define NV_PORT 			7650   
#define RF_PORT 			7651   


#define MAXMSG 				2048
#define MAXCLIENT 			40
#define MAXSRCREF  			5 

#define CONFBUF                         2000
#define MOTDBUF                         800
#define MSGBUF                          80

#define VIDEO 				1
#define CONTROL 			2
#define REF1VIDEO 			3
#define REF2VIDEO 			4
#define MAVEN				5
#define MAVEN_CNTL			6

#define ACKOPEN 			3
#define TIMEOUT 			10 
#define KALIVE 			        2 

#define MHEADERLEN                      10
#define NVVERSION                       0x40
#define NVCONTENT                       27
#define NVOBIT                          0x80
#define NVSBIT                          0x40
#define SDESC                           33 
#define FINAL                           0x80 

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif


typedef struct PacketLossData 
{
    unsigned short  packetsSent;
    unsigned short  packetsRecv;
} PacketLossData;


typedef struct addr
{
    short           family;        
    unsigned short  port;
    unsigned long   addr;
} addr;

typedef struct Routing {
    addr            dest;
    addr            src;
} Routing;

typedef struct VideoPacketHeader {
    Routing         routing;           /* 16 byte socket header for exploder */
    unsigned long   seqNum;
    short           message;
    short           dataType;
    short           len;
} VideoPacketHeader;

#define HEADERLEN 26

#define conferenceid routing.dest.port

typedef struct OpenContinueData 
{
        short			clientCount;
        unsigned long		seqNum;
        char			name[20];
        unsigned char		sendMode;
        unsigned char		recvMode;
        unsigned char		flags;
        unsigned char		version;
/*  
	ClientInfo		clients[clientCount] 
*/
} OpenContinueData;


#define REC_AUDIO		0x01
#define AUDIO_CAPABLE		0x02
#define WANT_LURCKERS		0x04
#define PC_CLIENT		0x08

#define VERSION1                1
#define VERSION5                5

typedef struct ClientInfo 
{
        unsigned long		clientIP;
        char			flags;
        char			na;
        char			IWillRecv;
        char			IWillSend;
        PacketLossData		loss;
} ClientInfo;

#define UPDATE_VIDEO  0x01
#define UPDATE_AUDIO  0x02
#define IWillRecAudio 0x04

typedef struct OpenContinuePacket 
{
        VideoPacketHeader	header;
        OpenContinueData	config;
} OpenContinuePacket;

typedef struct ControlPacket 
{
    VideoPacketHeader   header;
} ControlPacket;

typedef struct slist
{
   struct slist    *snd_nptr;
   struct client   *snd_client;
} slist;

typedef struct client 
{
    struct client     *clnt_nptr;           /* next client                                        */
    struct client     *clnt_pptr;           /* ptr to REF3_SOURCE if this client is a REF3_ORIGIN */
    addr              clnt_addr;            /* client's address                                   */
    slist             *clnt_vlist;          /* list of clients receiving video                    */
    slist             *clnt_alist;          /* list of clients receiving audio                    */
    short             clnt_rtimer;          /* client's recieve timer                             */
    short             clnt_stimer;          /* client's send timer                                */
    short             clnt_id;		    /* client's id                                        */	
    short             clnt_talker;          /* cnt of audio pkts rec'ed - used for logging current speaker */
    unsigned short    clnt_conf_id;         /* conference id the client connected with            */
    unsigned long     clnt_flags;           /* flags                                              */
    OpenContinueData  clnt_config;          /* client's configuration                             */
} client;

#define CLIENT        0x0001                /* normal Mac/PC client                                      */

#define BCC_CLIENT    0x0002                /* another reflector serving as a BCC client                 */
#define BCC_ORIGIN    0x0004                /* this client originates from a BCC server reflector        */
#define BCC_SERVER    0x0008                /* another reflector serving as a BCC server                 */

#define REF1_SERVER   0x0010                /* another reflector serving as a multicast source           */
#define REF1_ORIGIN   0x0020                /* this client originates from multicast source reflector (one way)    */
#define REF1_CLIENT   0x0040                /* another reflector serving as a multicast recipient        */


#define REF2_SERVER   0x0080                /* another reflector participating in a 2way mcast           */
#define REF2_ORIGIN   0x0100                /* this client originates from multicast group  reflector (two way)    */

#define SUSPEND       0x0200                /* this client is suspended                                  */
#define DELETE        0x0400                /* this client is deleted                                    */

#define REF3_SERVER   0x0800                /* another reflector participating in a 2way ucast           */
#define REF3_ORIGIN   0x1000                /* this client originates from unicat group reflector (two way)    */

typedef struct maven_client 
{

    struct maven_client   *mvn_nptr;           /* next client               */
    addr                  mvn_addr;            /* client's address          */
    unsigned long         mvn_seq;             /* sequence #                */
    short                 mvn_rtimer;          /* client's recieve timer    */
    short                 mvn_talker;          /* cnt of audio pkts rec'ed - used for logging current speaker */ 

} maven_client;

struct CtrlMsgHdr 
{
        unsigned char flags;
        unsigned char type;
        unsigned short confid;
        char idmsg[1];
};

#define CTRLMSGSIZE (4)

#define CTRL_TYPE_ID 1
#define CTRL_TYPE_DONE 2
#define CTRL_TYPE_IDLIST 3

