/*

Copyright 1993, 1994, Cornell University

Cornell hereby grants permission to use, copy, modify, and distribute this program for any purpose 
and without fee, provided that these copyright and permission notices appear on all copies and 
supporting documentation, the name of Cornell not be used in advertising or publicity pertaining 
to distribution of the program without specific prior permission, notice be given in supporting 
documentation that copying and distribution is by permission of Cornell.  CORNELL MAKES NO 
REPRESENTATIONS OR WARRANTEES, EXPRESS OR IMPLIED.  By way of example, but not limitation, 
CORNELL MAKES NO REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR 
PURPOSE OR THAT THE USE OF THIS SOFTWARE OR DOCUMENTATION WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS, 
TRADEMARKS, OR OTHER RIGHTS.  Cornell shall not be held liable for any liability with respect to 
any claim by the user or any other party arising from use of the program.

This material is partially based on work sponsored by the National Science Foundation under Cooperative 
Agreement No. NCR-9318337.  The government has certain rights in this material.

*/



#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <netinet/in.h>

#include "reflect.h"
#include "refmon.h"

#ifdef DEBUG
extern short debug;
#endif

extern client *find_client();
void          my_perror();

extern client *chead;
extern int vid_sock;

extern unsigned long bcc_server_list[MAXSRCREF];
extern short bcc_srv_cnt;

extern unsigned long bcc_client[MAXSRCREF];
extern short bcc_clt_cnt;

extern unsigned long ref_ucast_list[];
extern short ref_ucast_cnt;


struct sockaddr_in myaddr;                    

#ifdef MULTI
extern struct sockaddr_in nv_mcast;
extern int nv_mcast_sock;
extern struct sockaddr_in rfout_mcast;               
extern int rfout_mcast_sock;
extern struct sockaddr_in inout_mcast;               
extern int inout_mcast_sock;
#endif

/* 
   initialize the list of servers from whom we hope to attain a stream of 
   Blind Carbon Copy video
*/

open_bcc_servers()
{
    VideoPacketHeader pkt;
    client *ctmp,ctmp1;
    short cnt;

    for (cnt = 0; cnt < bcc_srv_cnt; cnt++)
    {
       pkt.routing.src.family = htons(kReflector);
       bcopy(&myaddr.sin_addr.s_addr,&pkt.routing.src.addr,4);
       pkt.routing.src.port = htons(VID_PORT);

       pkt.routing.dest.family = htons(kReflector);
       bcopy(&bcc_server_list[cnt],&pkt.routing.dest.addr,4);
       pkt.routing.dest.port = htons(VID_PORT);
   
       pkt.seqNum = 0;
       pkt.dataType = htons(kConfigRefType);
       pkt.len = htons(sizeof(VideoPacketHeader));
       pkt.message = htons(kOpenConnection);


#ifdef DEBUG
       if (debug)
          printf("sending BCC server keep alive ");
#endif

       /* 
	  if the server is already in our client list. use its actual client
	  ptr, else just us a temporary one for now
       */

       if ((ctmp = find_client(bcc_server_list[cnt])) == NULL)
       {
          bcopy(&bcc_server_list[cnt],&ctmp1.clnt_addr.addr,4);
	  ctmp1.clnt_addr.family = AF_INET;
	  ctmp1.clnt_addr.port = htons(VID_PORT);
	  ctmp1.clnt_flags = BCC_SERVER;
          write_pkt(&pkt,&ctmp1);
       }
       else
          write_pkt(&pkt,ctmp);
    }
}

open_ucast_ref()
{
    VideoPacketHeader pkt;
    client *ctmp,ctmp1;
    short cnt;

    for (cnt = 0; cnt < ref_ucast_cnt; cnt++)
    {
       pkt.routing.src.family = htons(kReflector);
       bcopy(&myaddr.sin_addr.s_addr,&pkt.routing.src.addr,4);
       pkt.routing.src.port = htons(VID_PORT);

       pkt.routing.dest.family = htons(kReflector);
       bcopy(&bcc_server_list[cnt],&pkt.routing.dest.addr,4);
       pkt.routing.dest.port = htons(VID_PORT);
   
       pkt.seqNum = 0;
       pkt.dataType = htons(kConfigRefType);
       pkt.len = htons(sizeof(VideoPacketHeader));
       pkt.message =  htons(kOpenConnection);


#ifdef DEBUG
       if (debug)
          printf("sending unicast server keep alive ");
#endif

       /* 
	  if the server is already in our client list. use its actual client
	  ptr, else just us a temporary one for now
       */

       if ((ctmp = find_client(ref_ucast_list[cnt])) == NULL)
       {
          bcopy(&ref_ucast_list[cnt],&ctmp1.clnt_addr.addr,4);
	  ctmp1.clnt_addr.family = AF_INET;
	  ctmp1.clnt_addr.port = htons(VID_PORT);
	  ctmp1.clnt_flags = REF3_SERVER;
          write_pkt(&pkt,&ctmp1);
       }
       else
          write_pkt(&pkt,ctmp);
    }
}

write_msg(clnt_addr,type,cptr)
    struct sockaddr_in  *clnt_addr;
    short               type;
    char                *cptr;
{
    static char         buf[MAXMSG];
    VideoPacketHeader   *pkt=(VideoPacketHeader *)buf;
    char                *tmp,clen;
    short               mlen,cnt;
    int                 len;

    pkt->routing.src.family = htons(kReflector);
    bcopy(&myaddr.sin_addr.s_addr,&pkt->routing.src.addr,4);
    pkt->routing.src.port = htons(VID_PORT);

    pkt->routing.dest.family = htons(kClient);
    bcopy(&clnt_addr->sin_addr,&pkt->routing.dest.addr,4);
    pkt->routing.dest.port = htons(VID_PORT);

    pkt->seqNum = 0;
    pkt->dataType = htons(type);
    pkt->message = 0;

    mlen = strlen(cptr);

    for (cnt = 0; cnt < mlen; cnt++)
    {

       if (*(cptr+cnt) == '\n')
          *(cptr+cnt) = '\r';
    }

    len = sizeof(VideoPacketHeader) + mlen + 2;

    pkt->len = htons(len);

    tmp = (char *) (pkt+1);

    if (type == kMessageType1)
    {
       len--;
       pkt->len = htons(len);
       clen = mlen;
       *tmp = clen;
       bcopy(cptr,(char *)(tmp+1),mlen);
    }
    else
    {
       bcopy(&mlen,tmp,2);
       bcopy(cptr,(char *)(tmp+2),mlen);
    }

#ifdef DEBUG
    if (debug)
       printf("sending reflector message %s to client at %s\n",cptr,inet_ntoa(clnt_addr->sin_addr));
#endif

    if (sendto(vid_sock,buf,len,0,clnt_addr,sizeof(struct sockaddr_in)) != len)
    {
       dolog("sendto error on old video socket");
    }
}

#ifdef MULTI
nv_mcast_wrt(vidptr,cltptr)
    VideoPacketHeader   *vidptr;
    client              *cltptr;

{
    unsigned char *cptr,*tmp;
    unsigned char id;
    unsigned short datatype;
    short len;

#ifdef DEBUG
       if (debug)
          printf("Multicasting (for NV) to %s\n",inet_ntoa(nv_mcast.sin_addr));
#endif

    cptr = ((unsigned char *) vidptr) + (HEADERLEN+2);
    cptr -= MHEADERLEN;
    tmp = cptr;

    len = ntohs(vidptr->len) - (HEADERLEN+2) + MHEADERLEN;
    datatype = vidptr->dataType;

    id = (cltptr->clnt_id % 32) + 32; 

    *tmp++ = (NVVERSION | id);

    *tmp = NVCONTENT;

    if (ntohs(vidptr->message) == kFrameEndMessage)
       *tmp |= NVSBIT;

    tmp++;

    bzero(tmp,6);
    tmp += 6;

    bcopy(&datatype,tmp,2);

    if (sendto(nv_mcast_sock,cptr,len,0,&nv_mcast,sizeof(struct sockaddr_in)) != len)
       dolog("mcast sendto error");

}


rf_mcast_wrt(vidptr)
    VideoPacketHeader   *vidptr;

{
#ifdef DEBUG
       if (debug)
          printf("Multicasting (for RF) to %s\n",inet_ntoa(rfout_mcast.sin_addr));
#endif

    if (sendto(rfout_mcast_sock,vidptr,ntohs(vidptr->len),0,&rfout_mcast,sizeof(struct sockaddr_in)) != ntohs(vidptr->len))
       dolog("mcast sendto error");

}

inout_mcast_wrt(vidptr)
    VideoPacketHeader   *vidptr;

{
#ifdef DEBUG
       if (debug)
          printf("Multicasting (to group) to %s\n",inet_ntoa(inout_mcast.sin_addr));
#endif

    if (sendto(inout_mcast_sock,vidptr,ntohs(vidptr->len),0,&inout_mcast,sizeof(struct sockaddr_in)) != ntohs(vidptr->len))
       dolog("mcast sendto error");

}


mcast_keep_alive(cltptr)
    client              *cltptr;
{

    unsigned char buf[100];
    unsigned char *cptr;
    unsigned char id;
    unsigned char len,blen;

#ifdef DEBUG
       if (debug)
          printf("Multicasting keep alive to %s for client %s\n",
		  inet_ntoa(nv_mcast.sin_addr),cltptr->clnt_config.name);
#endif

    len = strlen(cltptr->clnt_config.name) + 8;

    blen = len/4;

    if (blen != (len * 4))
       blen++;

    cptr = buf;

    id = (cltptr->clnt_id % 32) + 32; 

    *cptr++ = (NVVERSION | id);
    *cptr++ = NVCONTENT | NVOBIT;

    bzero(cptr,6);
    cptr += 6;

    *cptr++ = FINAL | SDESC;
    *cptr++ = blen;

    bcopy(&cltptr->clnt_addr.port,cptr,2);
    cptr += 2;

    bcopy(&cltptr->clnt_addr.addr,cptr,4);
    cptr += 4;

    strcpy(cptr,cltptr->clnt_config.name);
    
    cptr += strlen(cltptr->clnt_config.name);

    bzero(cptr,4);
    cptr += 4;

    len = (blen * 4) + 8;

    if (sendto(nv_mcast_sock,buf,len,0,&nv_mcast,sizeof(struct sockaddr_in)) != len)
       dolog("multicast keep alive sendto error");
}
#endif
