/*

Copyright 1993, 1994, Cornell University

Cornell hereby grants permission to use, copy, modify, and distribute this program for any purpose 
and without fee, provided that these copyright and permission notices appear on all copies and 
supporting documentation, the name of Cornell not be used in advertising or publicity pertaining 
to distribution of the program without specific prior permission, notice be given in supporting 
documentation that copying and distribution is by permission of Cornell.  CORNELL MAKES NO 
REPRESENTATIONS OR WARRANTEES, EXPRESS OR IMPLIED.  By way of example, but not limitation, 
CORNELL MAKES NO REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR 
PURPOSE OR THAT THE USE OF THIS SOFTWARE OR DOCUMENTATION WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS, 
TRADEMARKS, OR OTHER RIGHTS.  Cornell shall not be held liable for any liability with respect to 
any claim by the user or any other party arising from use of the program.

This material is partially based on work sponsored by the National Science Foundation under Cooperative 
Agreement No. NCR-9318337.  The government has certain rights in this material.

*/



#include <stdio.h>
#include <signal.h>
#include <sys/param.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <netinet/in.h>

#include <unistd.h>
#include <fcntl.h>

#include <sys/time.h>
#include <errno.h>
#include "reflect.h"

#ifdef DEBUG
extern short debug;
#endif


void                my_perror();

extern short                timer_expired;
extern int                  vid_sock,cntrl_sock,msg_sock;
extern int                  maven_sock,maven_cntl_sock,maven_port;
extern unsigned long        control_ip;

#ifdef MULTI
extern int                  nv_mcast_sock,rfin_mcast_sock,rfout_mcast_sock,inout_mcast_sock;
extern struct sockaddr_in   nv_mcast,rfin_mcast,rfout_mcast,inout_mcast;
extern int                  nv_ttl,rf_ttl,inout_ttl;
#endif

char                        hostname[MAXHOSTNAMELEN];


void init_socket()
{
    struct sockaddr_in  saddr;
#ifdef MULTI
    struct ip_mreq mreq;
#endif
    unsigned char ttl;
    unsigned char loop;


    if (maven_port != 0)
    {

       if ((maven_sock = socket(AF_INET, SOCK_DGRAM, 0))  < 0)
           my_perror("socket: maven socket");

       saddr.sin_family = AF_INET;
       saddr.sin_addr.s_addr = htonl(INADDR_ANY);
       saddr.sin_port = htons(maven_port);

       if (bind(maven_sock, (struct sockaddr *) &saddr, sizeof(saddr)) < 0)
           my_perror("bind error port");

       if ((maven_cntl_sock = socket(AF_INET, SOCK_DGRAM, 0))  < 0)
           my_perror("socket: maven control socket");

       saddr.sin_family = AF_INET;
       saddr.sin_addr.s_addr = htonl(INADDR_ANY);
       saddr.sin_port = htons(maven_port+1);

       if (bind(maven_cntl_sock, (struct sockaddr *) &saddr, sizeof(saddr)) < 0)
           my_perror("bind error port");

    }

    if ((vid_sock = socket(AF_INET, SOCK_DGRAM, 0))  < 0)
        my_perror("socket: video socket");

    saddr.sin_family = AF_INET;
    saddr.sin_addr.s_addr = htonl(INADDR_ANY);
    saddr.sin_port = htons(VID_PORT);

    if (bind(vid_sock, (struct sockaddr *) &saddr, sizeof(saddr)) < 0)
        my_perror("bind error port");

    if ((cntrl_sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
        my_perror("socket: control socket");

    saddr.sin_family = AF_INET;
    saddr.sin_addr.s_addr = INADDR_ANY;
    saddr.sin_port = htons(CONTROL_PORT);

    if (bind(cntrl_sock, &saddr, sizeof(struct sockaddr_in)) < 0)
        my_perror("bind error port");

#ifdef MULTI
    if (nv_mcast.sin_addr.s_addr)
    {
       if ((nv_mcast_sock = socket(AF_INET, SOCK_DGRAM, 0))  < 0)
           my_perror("socket: nv multicast socket");

       saddr.sin_family = AF_INET;
       saddr.sin_addr.s_addr = htonl(INADDR_ANY);
       saddr.sin_port = htons(NV_PORT);

       if (bind(nv_mcast_sock, (struct sockaddr *) &saddr, sizeof(saddr)) < 0)
           my_perror("bind error port");

       ttl = nv_ttl;
       setsockopt(nv_mcast_sock,IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl));
    }

    if (rfout_mcast.sin_addr.s_addr)
    {
       if ((rfout_mcast_sock = socket(AF_INET, SOCK_DGRAM, 0))  < 0)
           my_perror("socket: rfout multicast socket");

       saddr.sin_family = AF_INET;
       saddr.sin_addr.s_addr = htonl(INADDR_ANY);
       saddr.sin_port = htons(RF_PORT);

       if (bind(rfout_mcast_sock, (struct sockaddr *) &saddr, sizeof(saddr)) < 0)
           my_perror("bind error port");

       ttl = rf_ttl;
       setsockopt(rfout_mcast_sock,IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl));
    }

    if (rfin_mcast.sin_addr.s_addr)
    {
       if ((rfin_mcast_sock = socket(AF_INET, SOCK_DGRAM, 0))  < 0)
           my_perror("socket: rfin multicast socket");

       saddr.sin_family = AF_INET;
       saddr.sin_addr.s_addr = htonl(INADDR_ANY);
       saddr.sin_port = htons(RF_PORT);

       if (bind(rfin_mcast_sock, (struct sockaddr *) &saddr, sizeof(saddr)) < 0)
           my_perror("bind error port");

       mreq.imr_multiaddr.s_addr = rfin_mcast.sin_addr.s_addr;
       mreq.imr_interface.s_addr = INADDR_ANY;

       setsockopt(rfin_mcast_sock,IPPROTO_IP,IP_ADD_MEMBERSHIP,&mreq,sizeof(mreq));
    }

    if (inout_mcast.sin_addr.s_addr)
    {
       if ((inout_mcast_sock = socket(AF_INET, SOCK_DGRAM, 0))  < 0)
           my_perror("socket: inout multicast socket");

       saddr.sin_family = AF_INET;
       saddr.sin_addr.s_addr = htonl(INADDR_ANY);
       saddr.sin_port = htons(RF_PORT);

       if (bind(inout_mcast_sock, (struct sockaddr *) &saddr, sizeof(saddr)) < 0)
           my_perror("bind error port");

       ttl = inout_ttl;
       setsockopt(inout_mcast_sock,IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl));

       mreq.imr_multiaddr.s_addr = inout_mcast.sin_addr.s_addr;
       mreq.imr_interface.s_addr = INADDR_ANY;

       setsockopt(inout_mcast_sock,IPPROTO_IP,IP_ADD_MEMBERSHIP,&mreq,sizeof(mreq));

       loop = 0;
       setsockopt(inout_mcast_sock,IPPROTO_IP,IP_MULTICAST_LOOP,&loop,sizeof(loop));


    }

#endif 

    if ((fcntl(vid_sock, F_SETFL, FNDELAY)) < 0 )
        my_perror("fcntl error");

    if (maven_sock)
    {
       if ((fcntl(maven_sock, F_SETFL, FNDELAY)) < 0 )
           my_perror("fcntl error");

       if ((fcntl(maven_cntl_sock, F_SETFL, FNDELAY)) < 0 )
           my_perror("fcntl error");
    }

    if ((fcntl(cntrl_sock, F_SETFL, FNDELAY)) < 0) 
        my_perror("fcntl error");

    if (gethostname(hostname,MAXHOSTNAMELEN) < 0)
        my_perror("getnostname");

    msg_sock = 0;
    setsockopt(cntrl_sock,SOL_SOCKET,SO_REUSEADDR,1,0);
    listen(cntrl_sock,5);
}

void init_timer()
{
    struct itimerval    value;
    struct itimerval    ovalue;
    void                sighandler();

    value.it_interval.tv_sec = 1;
    value.it_interval.tv_usec = 0;
    value.it_value.tv_sec = 1;
    value.it_value.tv_usec = 0;

    setitimer(ITIMER_REAL, &value, &ovalue);

    signal(SIGALRM, sighandler);
}

void sighandler()
{
    timer_expired++;
    signal(SIGALRM, sighandler);
}

int receive(pkt,msglen,caddr)
    char               *pkt;
    int                *msglen;
    struct sockaddr_in *caddr; 
{
    fd_set                      readfds;
    int                         cadrlen,ready;

    while (1)
    {
       if (timer_expired)
       {
          timer_expired--;
          do_timer();
       }

       FD_ZERO(&readfds);
       FD_SET(vid_sock, &readfds);

       if (maven_sock)
       {
          FD_SET(maven_sock,&readfds);
          FD_SET(maven_cntl_sock,&readfds);
       }

#ifdef MULTI
       if (rfin_mcast.sin_addr.s_addr)
          FD_SET(rfin_mcast_sock,&readfds);
       
       if (inout_mcast.sin_addr.s_addr)
          FD_SET(inout_mcast_sock,&readfds);
#endif
       if (msg_sock == 0)
          FD_SET(cntrl_sock, &readfds);

       if (msg_sock != 0)
          FD_SET(msg_sock, &readfds);
       
       if ((ready = select(FD_SETSIZE, &readfds, NULL, NULL, NULL)) < 0)
       {
          if (errno == EINTR)
             continue;
          else
             my_perror("select error");
       }

       if (ready == 0)
           continue;

       cadrlen = sizeof(struct sockaddr_in);

       if (FD_ISSET(vid_sock, &readfds)) 
       {
          
          if ((*msglen = recvfrom(vid_sock,pkt,MAXMSG,0,caddr,&cadrlen)) < 0)
          {
             if (errno == EINTR)
                continue;

             my_perror("recvfrom error on vid_sock");
          }

          return(VIDEO);
       }

       if (FD_ISSET(maven_sock, &readfds)) 
       {
          
          if ((*msglen = recvfrom(maven_sock,pkt,MAXMSG,0,caddr,&cadrlen)) < 0)
          {
             if (errno == EINTR)
                continue;

             my_perror("recvfrom error on maven_sock");
          }

          return(MAVEN);
       }

       if (FD_ISSET(maven_cntl_sock, &readfds)) 
       {
          
          if ((*msglen = recvfrom(maven_cntl_sock,pkt,MAXMSG,0,caddr,&cadrlen)) < 0)
          {
             if (errno == EINTR)
                continue;

             my_perror("recvfrom error on maven_cntl_sock");
          }

          return(MAVEN_CNTL);
       }

#ifdef MULTI
       if ((rfin_mcast_sock) && (FD_ISSET(rfin_mcast_sock, &readfds)))
       {
          
          if ((*msglen = recvfrom(rfin_mcast_sock,pkt,MAXMSG,0,caddr,&cadrlen)) < 0)
          {
             if (errno == EINTR)
                continue;

             my_perror("recvfrom error on refin_mcast_sock");
          }

          return(REF1VIDEO);
       }

       if ((inout_mcast_sock) && (FD_ISSET(inout_mcast_sock, &readfds)))
       {
          
          if ((*msglen = recvfrom(inout_mcast_sock,pkt,MAXMSG,0,caddr,&cadrlen)) < 0)
          {
             if (errno == EINTR)
                continue;

             my_perror("recvfrom error on inout_mcast_sock");
          }

          return(REF2VIDEO);
       }
#endif

       if (FD_ISSET(cntrl_sock, &readfds))
       {
          if (msg_sock != 0)
          {
             dolog("control message with a non-zero msg_sock\n");
	     exit(1);
	  }

          if ((msg_sock = accept(cntrl_sock,caddr,&cadrlen)) < 0)
          {
             if (errno == EINTR)
             {
	        msg_sock = 0;
                continue;
             }

             my_perror("accept");
             exit(1);
          }


          if ((control_ip == 0) || ((control_ip != -1) && (control_ip != caddr->sin_addr.s_addr)))
          {
             dolog("Control message from %s is disallowed\n",inet_ntoa(caddr->sin_addr));
             close(msg_sock);
             msg_sock = 0;
             continue;
          }


          dolog("incoming control message new msg_sock is %d\n",msg_sock);
          setsockopt(msg_sock,SOL_SOCKET,SO_REUSEADDR,1,0);
	  continue;
        }

       if (FD_ISSET(msg_sock, &readfds))
       {
          if ((*msglen = recv(msg_sock,pkt,MAXMSG,0)) < 0)
          {
             if (errno == EINTR)
                continue;

             my_perror("recvfrom error on msg_sock");
             exit (1);
          }
   
          if (*msglen == 0)
          {
             close(msg_sock);
             msg_sock = 0;
             continue;
          }
   
          return (CONTROL);
       }
    }
}
