/*

Copyright 1993, 1994, Cornell University

Cornell hereby grants permission to use, copy, modify, and distribute this program for any purpose 
and without fee, provided that these copyright and permission notices appear on all copies and 
supporting documentation, the name of Cornell not be used in advertising or publicity pertaining 
to distribution of the program without specific prior permission, notice be given in supporting 
documentation that copying and distribution is by permission of Cornell.  CORNELL MAKES NO 
REPRESENTATIONS OR WARRANTEES, EXPRESS OR IMPLIED.  By way of example, but not limitation, 
CORNELL MAKES NO REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR 
PURPOSE OR THAT THE USE OF THIS SOFTWARE OR DOCUMENTATION WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS, 
TRADEMARKS, OR OTHER RIGHTS.  Cornell shall not be held liable for any liability with respect to 
any claim by the user or any other party arising from use of the program.

This material is partially based on work sponsored by the National Science Foundation under Cooperative 
Agreement No. NCR-9318337.  The government has certain rights in this material.

*/



#include <stdio.h>
#include <varargs.h>
#include <signal.h>
#include <errno.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/time.h>
#include <net/route.h>
#include <net/if.h>
#include <netinet/in.h>

#include "reflect.h"
#include "refmon.h"

#ifdef DEBUG
extern short debug;
#endif

void                my_perror();
short               get_client_id();
extern              char *ctime();

extern int vid_sock;
extern int maven_port;
extern int maven_sock;
extern int maven_cntl_sock;

extern client *chead;
extern maven_client *mhead;
extern short self_reflect;
extern struct sockaddr_in myaddr; 

extern unsigned long restrict_list[];
extern short restrict_cnt;          

extern unsigned long deny_list[];
extern short deny_cnt;

extern int log_limit;

extern unsigned long bcc_server_list[];     
extern short bcc_srv_cnt;                  

extern unsigned long bcc_client_list[];   
extern short bcc_clt_cnt;                

extern unsigned long ref_ucast_list[]; 
extern short ref_ucast_cnt;     

extern int maxallowed;

extern char motd_buf[];
extern short motd_len;

extern char mp_buf[];

extern short local_senders;

#ifdef MULTI
extern struct sockaddr_in nv_mcast;               
extern int nv_mcast_sock;

extern struct sockaddr_in rfout_mcast;             
extern int rfout_mcast_sock;

extern struct sockaddr_in rfin_mcast;               
extern int rfin_mcast_sock;

extern struct sockaddr_in inout_mcast;               
extern int inout_mcast_sock;

#endif

unsigned short client_cnt;

extern unsigned short bcc_client_cnt;
extern unsigned short bcc_origin_cnt;
extern unsigned short bcc_server_cnt;

extern unsigned short ref1_server_cnt;
extern unsigned short ref1_origin_cnt;
extern unsigned short ref1_client_cnt;

extern unsigned short ref2_server_cnt;
extern unsigned short ref2_origin_cnt;

extern unsigned short ref3_server_cnt;
extern unsigned short ref3_origin_cnt;


extern char *tracefile;                              

extern FILE *log_file;

#define MAXIF 20        /* Maximum number of interfaces expected        */

#if     BSD > 43
#define socksize(x)     ((struct sockaddr *)x)->sa_len
#else
#define socksize(x)     sizeof *(x)
#endif

slist *unlink_slist();
client *unlink_client();

short get_client_id()
{
    client *ctmp;
    short  client_id,id_not_set;

    client_id = 1;
    id_not_set = 1;

    /* find the next free client id  */

    while (id_not_set)
    {
       ctmp = chead;
       while (ctmp != NULL)
       {
         if (ctmp->clnt_id == client_id)
         {
            client_id++;
            break;
         }
         ctmp = ctmp->clnt_nptr;
       }
       if (ctmp == NULL)
         id_not_set = 0;
    }

    return(client_id);
}

client *find_client(caddr) 
     unsigned long caddr;
{
    client *ctmp;

    ctmp = chead;

    while (ctmp != NULL)
    {
      if (ctmp->clnt_addr.addr == caddr)
         return(ctmp);
      ctmp = ctmp->clnt_nptr;
    }
    return(NULL);
}

maven_client *find_maven_client(caddr) 
    unsigned long caddr;
{
    maven_client *ctmp;

    ctmp = mhead;

    while (ctmp != NULL)
    {
      if (ctmp->mvn_addr.addr == caddr)
         return(ctmp);
      ctmp = ctmp->mvn_nptr;
    }
    return(NULL);
}


maven_client *new_maven_client(caddr) 
    unsigned long caddr;
{
    maven_client *ctmp;

    if ((ctmp = (maven_client *) calloc(1,sizeof(maven_client))) == NULL)
       return(NULL);

    ctmp->mvn_addr.addr = caddr;           
    ctmp->mvn_addr.port = maven_port;           
    ctmp->mvn_addr.family = AF_INET;           

    ctmp->mvn_nptr = mhead;
    mhead = ctmp;
    return(ctmp);
}

client *open_connection(vidptr,csock,type)
   VideoPacketHeader   *vidptr;
   struct sockaddr_in  *csock;
   short               type;
{
    client        *cltptr,*ctmp;
    char          *cptr;
    short         client_id,cnt;

    switch (type)
    {
       case BCC_CLIENT:
          dolog("BCC client at %s is opening connection\n", inet_ntoa(csock->sin_addr));

          cltptr = (client *) get_client();
          cltptr->clnt_addr.family = AF_INET;
          cltptr->clnt_addr.port = vidptr->routing.src.port;
          cltptr->clnt_flags |= BCC_CLIENT; 
          bcopy(&csock->sin_addr,&cltptr->clnt_addr.addr,4);
          strcpy(cltptr->clnt_config.name,"BCC CLIENT");
	  
          cltptr->clnt_nptr = chead;
          chead = cltptr;
	  bcc_client_cnt++;
          return(cltptr);

       case BCC_SERVER:
          dolog("BCC server at %s is opening a connection\n", inet_ntoa(csock->sin_addr));

          cltptr = (client *) get_client();
          cltptr->clnt_addr.family = AF_INET;
          cltptr->clnt_addr.port = vidptr->routing.src.port;
          cltptr->clnt_flags |= BCC_SERVER;
          bcopy(&csock->sin_addr,&cltptr->clnt_addr.addr,4);
          strcpy(cltptr->clnt_config.name,"BCC SERVER");

          cltptr->clnt_nptr = chead;
          chead = cltptr;

	  bcc_server_cnt++;

          open_connection(vidptr,(struct sockaddr_in *) &vidptr->routing.src,BCC_ORIGIN);
          return(cltptr);

       case BCC_ORIGIN:  
          dolog("BCC server's client at %s is opening a connection\n", inet_ntoa(csock->sin_addr));

          if ((client_id = get_client_id()) == -1)
          {
             dolog("maximum # of clients exceeded\n");
             return(NULL);
          }
      
          cltptr = (client *) get_client();
          cltptr->clnt_addr.family = AF_INET;
          cltptr->clnt_addr.port = vidptr->routing.src.port;
          cltptr->clnt_flags |= BCC_ORIGIN;
          cltptr->clnt_id = client_id;
          bcopy(&csock->sin_addr,&cltptr->clnt_addr.addr,4);
      
          cltptr->clnt_nptr = chead;
          chead = cltptr;

	  bcc_origin_cnt++;
      
          cptr = ((char *) vidptr + HEADERLEN);
          update_client(cltptr,cptr,FALSE,FALSE);

          return(cltptr);

       case REF1_CLIENT:   
          dolog("REF downstream client at %s is opening connection\n", inet_ntoa(csock->sin_addr));

          cltptr = (client *) get_client();
          cltptr->clnt_addr.family = AF_INET;
          cltptr->clnt_addr.port = vidptr->routing.src.port;
          cltptr->clnt_flags |= REF1_CLIENT; 
          bcopy(&csock->sin_addr,&cltptr->clnt_addr.addr,4);
          strcpy(cltptr->clnt_config.name,"Downsteam REF");

          cltptr->clnt_nptr = chead;
          chead = cltptr;

	  ref1_client_cnt++;

          return(cltptr);

       case REF1_SERVER:

          dolog("REF upstream server at %s is opening a connection\n", inet_ntoa(csock->sin_addr));

          cltptr = (client *) get_client();
          cltptr->clnt_addr.family = AF_INET;
          cltptr->clnt_addr.port = vidptr->routing.src.port;
          cltptr->clnt_flags |= REF1_SERVER;
          bcopy(&csock->sin_addr,&cltptr->clnt_addr.addr,4);
          strcpy(cltptr->clnt_config.name,"Upstream REF");

          cltptr->clnt_nptr = chead;
          chead = cltptr;

          ref1_server_cnt++;

          open_connection(vidptr,(struct sockaddr_in *) &vidptr->routing.src,REF1_ORIGIN);
          return(cltptr);

       case REF1_ORIGIN:   
          dolog("REF upstream server's client at %s is opening a connection\n", inet_ntoa(csock->sin_addr));

          if ((client_id = get_client_id()) == -1)
          {
             dolog("maximum # of clients exceeded\n");
             return(NULL);
          }
      
          cltptr = (client *) get_client();
          cltptr->clnt_addr.family = AF_INET;
          cltptr->clnt_addr.port = vidptr->routing.src.port;
          cltptr->clnt_flags |= REF1_ORIGIN;
          cltptr->clnt_id = client_id;
          bcopy(&csock->sin_addr,&cltptr->clnt_addr.addr,4);
      
          cltptr->clnt_nptr = chead;
          chead = cltptr;

	  ref1_origin_cnt++;

          cptr = ((char *) vidptr + HEADERLEN);
          update_client(cltptr,cptr,FALSE,FALSE);

          return(cltptr);

       case REF2_SERVER:   

          dolog("REF group server at %s is opening a connection\n", inet_ntoa(csock->sin_addr));

          cltptr = (client *) get_client();
          cltptr->clnt_addr.family = AF_INET;
          cltptr->clnt_addr.port = vidptr->routing.src.port;
          cltptr->clnt_flags |= REF2_SERVER;
          bcopy(&csock->sin_addr,&cltptr->clnt_addr.addr,4);
          strcpy(cltptr->clnt_config.name,"Group REF");

          cltptr->clnt_nptr = chead;
          chead = cltptr;

	  ref2_server_cnt++;

          open_connection(vidptr,(struct sockaddr_in *) &vidptr->routing.src,REF2_ORIGIN);
          return(cltptr);

       case REF2_ORIGIN:

          dolog("REF group server's client at %s is opening a connection\n", inet_ntoa(csock->sin_addr));

          if ((client_id = get_client_id()) == -1)
          {
             dolog("maximum # of clients exceeded\n");
             return(NULL);
          }

          cltptr = (client *) get_client();
          cltptr->clnt_addr.family = AF_INET;
          cltptr->clnt_addr.port = vidptr->routing.src.port;
          cltptr->clnt_flags |= REF2_ORIGIN;
          cltptr->clnt_id = client_id;
          bcopy(&csock->sin_addr,&cltptr->clnt_addr.addr,4);

          cltptr->clnt_nptr = chead;
          chead = cltptr;
      
	  ref2_origin_cnt++;

          cptr = ((char *) vidptr + HEADERLEN);
          update_client(cltptr,cptr,FALSE,FALSE);
          return(cltptr);

       case REF3_SERVER:   

          dolog("REF unicast group server at %s is opening a connection\n", inet_ntoa(csock->sin_addr));

          cltptr = (client *) get_client();
          cltptr->clnt_addr.family = AF_INET;
          cltptr->clnt_addr.port = vidptr->routing.src.port;
          cltptr->clnt_flags |= REF3_SERVER;
          bcopy(&csock->sin_addr,&cltptr->clnt_addr.addr,4);
          strcpy(cltptr->clnt_config.name,"Group Uni REF");

          for (cnt=0; cnt<ref_ucast_cnt; cnt++)
             if (ref_ucast_list[cnt] == csock->sin_addr.s_addr)
	     {
                cltptr->clnt_id = cnt;
		break;
	     } 

          cltptr->clnt_nptr = chead;
          chead = cltptr;

	  ref3_server_cnt++;

          return(cltptr);

       case REF3_ORIGIN:

          dolog("REF unicast group server's client at %s is opening a connection\n", inet_ntoa(csock->sin_addr));

          if ((client_id = get_client_id()) == -1)
          {
             dolog("maximum # of clients exceeded\n");
             return(NULL);
          }

          cltptr = (client *) get_client();
          cltptr->clnt_addr.family = AF_INET;
          cltptr->clnt_addr.port = vidptr->routing.src.port;
          cltptr->clnt_flags |= REF3_ORIGIN;
          cltptr->clnt_id = client_id;
          bcopy(&csock->sin_addr,&cltptr->clnt_addr.addr,4);

          cltptr->clnt_nptr = chead;
          chead = cltptr;
      
	  ref3_origin_cnt++;

          cptr = ((char *) vidptr + HEADERLEN);
          update_client(cltptr,cptr,FALSE,FALSE);
          return(cltptr);


       case CLIENT:
          dolog("Client at %s is opening a connection\n", inet_ntoa(csock->sin_addr));

          if ((client_id = get_client_id()) == -1)
          {
             dolog("maximum # of clients exceeded\n");
             return(NULL);
          }
      
          cltptr = (client *) get_client();
          cltptr->clnt_addr.family = AF_INET;
          cltptr->clnt_addr.port = vidptr->routing.src.port;
          cltptr->clnt_flags |= CLIENT;
          cltptr->clnt_id = client_id;
	  cltptr->clnt_conf_id = vidptr->conferenceid;
          bcopy(&csock->sin_addr,&cltptr->clnt_addr.addr,4);
      
          cltptr->clnt_nptr = chead;
          chead = cltptr;

	  client_cnt++;
      
          cptr = ((char *) vidptr + HEADERLEN);
          update_client(cltptr,cptr,TRUE,TRUE);

	  if (client_cnt >= maxallowed)
	  {
             write_msg(&cltptr->clnt_addr,kMessageType1,mp_buf);
	     delete_client(cltptr);
             return(NULL);
	  }

	  if (motd_len != 0)
	     write_msg(&cltptr->clnt_addr,kMessageType2,motd_buf);

          return(cltptr);
    }
}

continue_connection(cltptr,vidptr)
   client              *cltptr;
   VideoPacketHeader   *vidptr;
{
    client        *ctmp;
    char          *cptr;
    unsigned long seqnum,diff;
    short         type;
    char          doslist;

    if (cltptr->clnt_flags & CLIENT)
    {
       cptr = ((char *)vidptr + HEADERLEN);
       seqnum = ntohl(*((unsigned long *) (cptr + 2)));
       
       diff = seqnum - cltptr->clnt_config.seqNum;
       if (diff == 1)
          update_client(cltptr,cptr,TRUE,FALSE);
       else
	  if (diff > 1)
             update_client(cltptr,cptr,TRUE,TRUE);

       if (cltptr->clnt_stimer >= ACKOPEN)
       {
#ifdef DEBUG
          if (debug)
             printf("ACK timeout sending open ack to %s\n",cltptr->clnt_config.name);
#endif
          write_cmsg(kGotOpen,cltptr);
       }
       return;
    }


    if (cltptr->clnt_flags & BCC_SERVER)
    {
       type = BCC_ORIGIN;
       doslist = FALSE;
    }
    else
    if (cltptr->clnt_flags & REF1_SERVER)
    {
       type = REF1_ORIGIN;
       doslist = FALSE;
    }
    else
    if (cltptr->clnt_flags & REF2_SERVER)
    {
       type = REF2_ORIGIN;
       doslist = TRUE;
    }
    else
    if (cltptr->clnt_flags & REF3_SERVER)
    {
       type = REF3_ORIGIN;
       doslist = TRUE;
    }
    else
       return;


    if ((ctmp = find_client(vidptr->routing.src.addr)) != NULL)
    {
       ctmp->clnt_rtimer = 0;
       ctmp->clnt_flags &= ~SUSPEND;

       cptr = ((char *)vidptr + HEADERLEN);
       seqnum = ntohl(*((unsigned long *) (cptr + 2)));
       
       diff = seqnum - ctmp->clnt_config.seqNum;
       if (diff == 1)
          update_client(ctmp,cptr,doslist,FALSE);
       else
	  if (diff > 1)
             update_client(ctmp,cptr,doslist,TRUE);
    }
    else
       if ((ctmp = open_connection(vidptr,(struct sockaddr_in *) &vidptr->routing.src,type)) != NULL)
       {
          if (type == REF3_ORIGIN)
	     ctmp->clnt_pptr = cltptr;
       }

     
}

delete_client(cltptr)
   client *cltptr;
{
    client *ctmp;
    slist  *stmp,*stmp1;

    dolog("deleting client %s\n",cltptr->clnt_config.name);

    chead = unlink_client(chead,cltptr);

    if ((cltptr->clnt_flags & DELETE) == 0)
    {
       dolog("delete_client: client not found\n");
       exit(-1);
    }

    ctmp = chead;
    while (ctmp != NULL)
    {
       ctmp->clnt_vlist = unlink_slist(ctmp->clnt_vlist,cltptr);
       ctmp->clnt_alist = unlink_slist(ctmp->clnt_alist,cltptr);
       ctmp = ctmp->clnt_nptr;
    }

    stmp = cltptr->clnt_vlist;
    while (stmp != NULL)
    {
       stmp1 = stmp;
       stmp = stmp->snd_nptr;
       free_slist(stmp1);
    }

    stmp = cltptr->clnt_alist;
    while (stmp != NULL)
    {
       stmp1 = stmp;
       stmp = stmp->snd_nptr;
       free_slist(stmp1);
    }

    if (cltptr->clnt_flags & CLIENT)
       client_cnt--;
    else
    if (cltptr->clnt_flags & BCC_CLIENT)
       bcc_client_cnt--;
    else
    if (cltptr->clnt_flags & BCC_SERVER)
       bcc_client_cnt--;
    else
    if (cltptr->clnt_flags & BCC_ORIGIN)
       bcc_client_cnt--;
    else
    if (cltptr->clnt_flags & REF1_CLIENT)
       ref1_client_cnt--;
    else
    if (cltptr->clnt_flags & REF1_SERVER)
       ref1_server_cnt--;
    else
    if (cltptr->clnt_flags & REF1_ORIGIN)
       ref1_origin_cnt--;
    else
    if (cltptr->clnt_flags & REF2_SERVER)
       ref2_server_cnt--;
    else
    if (cltptr->clnt_flags & REF2_ORIGIN)
       ref2_origin_cnt--;
    else
    if (cltptr->clnt_flags & REF3_SERVER)
       ref3_server_cnt--;
    else
    if (cltptr->clnt_flags & REF3_ORIGIN)
       ref3_origin_cnt--;
    else
    {
       dolog("delete_client: unknown type\n");
       exit(-1);
    }

    free_client(cltptr);
}

write_cmsg(cmd,cltptr)
    short	cmd;
    client      *cltptr;

{
    VideoPacketHeader pkt;

    pkt.routing.src.family = htons(kReflector);
    bcopy(&myaddr.sin_addr.s_addr,&pkt.routing.src.addr,4);
    pkt.routing.src.addr = 0;
    pkt.routing.src.port = htons(VID_PORT);

    pkt.routing.dest.family = htons(kClient);
    bcopy(&cltptr->clnt_addr.addr,&pkt.routing.dest.addr,4);
    pkt.routing.dest.port = htons(VID_PORT);

    pkt.seqNum = 0;
    pkt.dataType = htons(kControlType);
    pkt.len = htons(sizeof(VideoPacketHeader));
    pkt.message = htons(cmd);

    write_pkt(&pkt,cltptr);
}


write_pkt(vidptr,cltptr)
    VideoPacketHeader        *vidptr;
    client                   *cltptr;
{
    struct in_addr in;

    if (cltptr->clnt_flags & REF3_ORIGIN)
       cltptr = cltptr->clnt_pptr;


#ifdef DEBUG
    if (debug)
    {
       in.s_addr = cltptr->clnt_addr.addr;
       printf("sending to %s\n",inet_ntoa(in));
    }
#endif

    cltptr->clnt_stimer = 0;


    if (sendto(vid_sock,vidptr,ntohs(vidptr->len),0,&cltptr->clnt_addr,sizeof(struct sockaddr_in)) != ntohs(vidptr->len))
       dolog("sendto error %d client %s\n",errno,cltptr->clnt_config.name);
}

update_client(cltptr,cptr,doslist,all)
    client           *cltptr;
    char             *cptr;
    char             doslist;
    char             all;
{
    ClientInfo       *ciptr;
    client           *ctmp;
    slist            *stmp;
    unsigned long    old_seq;

    char             name_len;
    short            cnt;
    struct in_addr   in;


    cltptr->clnt_config.clientCount = ntohs(*((short *)cptr));
    cptr += sizeof(short);

    if (cltptr->clnt_config.clientCount > 20)
       return;

    dolog("updating client %s client count %d   seq %ld\n",cltptr->clnt_config.name,cltptr->clnt_config.clientCount,cltptr->clnt_config.seqNum);

    old_seq = cltptr->clnt_config.seqNum;
    cltptr->clnt_config.seqNum = ntohl(*((unsigned long *)cptr));
    cptr += sizeof(unsigned long);

    name_len = *cptr++; 
    if (name_len > 19)
    {
       dolog("bogus name length %d\n",name_len);
       name_len = 19;
    }

    for (cnt = 0; cnt < name_len; cnt++)
    {
       cltptr->clnt_config.name[cnt] = cptr[cnt];

       if (cltptr->clnt_config.name[cnt] == '\n')
          cltptr->clnt_config.name[cnt] = ' ';

       if (cltptr->clnt_config.name[cnt] == '\r')
          cltptr->clnt_config.name[cnt] = ' ';
    }       

    cltptr->clnt_config.name[name_len] = 0;

    cptr += 19;

    if ((cltptr->clnt_flags & CLIENT) && (local_senders == 0))
       *cptr = 0;

    cltptr->clnt_config.sendMode = *cptr++;

    cltptr->clnt_config.recvMode = *cptr++;
    cltptr->clnt_config.flags = *cptr++;
    if ((cltptr->clnt_config.version = *cptr++) < VERSION1)
       all = TRUE;

    /* delete this client from the other client's send list                                     */

    if (doslist == FALSE)
       return;

    if (all)
    {
       ctmp = chead;
       while (ctmp != NULL)
       {
          ctmp->clnt_vlist = unlink_slist(ctmp->clnt_vlist,cltptr);
          ctmp->clnt_alist = unlink_slist(ctmp->clnt_alist,cltptr);
          ctmp = ctmp->clnt_nptr;
       }
    }

    /*
    ctmp = chead;
    while (ctmp != NULL)
    {
       stmp = ctmp->clnt_vlist;
       cnt = 0;
       while (stmp != NULL)
       {
	  cnt++;
          if (stmp->snd_client == cltptr)
	     dolog("SLIST IS DAMAGED\n");
	  stmp = stmp->snd_nptr;
       }
       dolog("%s's slist count %d\n",ctmp->clnt_config.name,cnt);
       ctmp = ctmp->clnt_nptr;
    }
    */

    ciptr = (ClientInfo *) (cptr);

    for (cnt = 0; cnt < cltptr->clnt_config.clientCount; cnt++)
    {
       if (all)
       {
          if (ciptr->IWillRecv)
          {
             if ((ctmp = find_client(ciptr->clientIP,chead)) == NULL)
             {
                in.s_addr = ciptr->clientIP;
                dolog("%s wants to recieve video from %s but client not found\n",cltptr->clnt_config.name,inet_ntoa(in));
                cltptr->clnt_config.seqNum = old_seq;
             }
             else
             {
                if (cltptr->clnt_config.version >= VERSION5)
                   dolog("%s wants to recieve video from %s\n",cltptr->clnt_config.name,ctmp->clnt_config.name);
		else
                   dolog("%s wants to recieve audio and video from %s\n",cltptr->clnt_config.name,ctmp->clnt_config.name);

		stmp = (slist *) get_slist();
                stmp->snd_nptr = ctmp->clnt_vlist;
                ctmp->clnt_vlist = stmp;
                stmp->snd_client = cltptr;

                if (cltptr->clnt_config.version  < VERSION5)
		{
		   stmp = (slist *) get_slist();
                   stmp->snd_nptr = ctmp->clnt_alist;
                   ctmp->clnt_alist = stmp;
                   stmp->snd_client = cltptr;
                }

             }
          }

	  if (ciptr->flags & IWillRecAudio)
	  {
             if ((ctmp = find_client(ciptr->clientIP,chead)) == NULL)
             {
                in.s_addr = ciptr->clientIP;
                dolog("%s wants to recieve audio from %s but client not found\n",cltptr->clnt_config.name,inet_ntoa(in));
                cltptr->clnt_config.seqNum = old_seq;
             }
             else
             {
                dolog("%s wants to recieve audio from %s\n",cltptr->clnt_config.name,ctmp->clnt_config.name);
  
		stmp = (slist *) get_slist();
                stmp->snd_nptr = ctmp->clnt_alist;
                ctmp->clnt_alist = stmp;
                stmp->snd_client = cltptr;
             }
	  }

       }	
       else
       {
	  if (ciptr->flags & UPDATE_VIDEO)
	  {
             if ((ctmp = find_client(ciptr->clientIP,chead)) == NULL)
             {
                in.s_addr = ciptr->clientIP;
                dolog("%s wants to update client at %s but client not found\n",cltptr->clnt_config.name,inet_ntoa(in));
                cltptr->clnt_config.seqNum = old_seq;
             }
	     else
	     {
                if (ciptr->IWillRecv)
		{
                   if (cltptr->clnt_config.version  >= VERSION5)
                      dolog("%s wants to recieve video from %s\n",cltptr->clnt_config.name,ctmp->clnt_config.name);
		   else
                      dolog("%s wants to recieve video and audio from %s\n",cltptr->clnt_config.name,ctmp->clnt_config.name);

		   stmp = (slist *) get_slist();
                   stmp->snd_nptr = ctmp->clnt_vlist;
                   ctmp->clnt_vlist = stmp;
                   stmp->snd_client = cltptr;

                   if (cltptr->clnt_config.version  < VERSION5)
		   {
		      stmp = (slist *) get_slist();
                      stmp->snd_nptr = ctmp->clnt_alist;
                      ctmp->clnt_alist = stmp;
                      stmp->snd_client = cltptr;
		   }
		}
		else
		{
                   if (cltptr->clnt_config.version  >= VERSION5)
		   {
                      dolog("%s no longer wants to recieve video from %s\n",cltptr->clnt_config.name,ctmp->clnt_config.name);
                      ctmp->clnt_vlist = unlink_slist(ctmp->clnt_vlist,cltptr);
                   }
		   else
		   {
                      dolog("%s no longer wants to recieve video or audio from %s\n",cltptr->clnt_config.name,ctmp->clnt_config.name);
                      ctmp->clnt_vlist = unlink_slist(ctmp->clnt_vlist,cltptr);
                      ctmp->clnt_alist = unlink_slist(ctmp->clnt_alist,cltptr);
                   }
                }
	     }
	  }

	  if (ciptr->flags & UPDATE_AUDIO)
	  {
             if ((ctmp = find_client(ciptr->clientIP,chead)) == NULL)
             {
                in.s_addr = ciptr->clientIP;
                dolog("%s wants to update client at %s but client not found\n",cltptr->clnt_config.name,inet_ntoa(in));
                cltptr->clnt_config.seqNum = old_seq;
             }
	     else
	     {
                if (ciptr->flags & IWillRecAudio)
		{
                   dolog("%s wants to recieve audio from %s\n",cltptr->clnt_config.name,ctmp->clnt_config.name);
   
		   stmp = (slist *) get_slist();
                   stmp->snd_nptr = ctmp->clnt_alist;
                   ctmp->clnt_alist = stmp;
                   stmp->snd_client = cltptr;
		}
		else
		{
                   dolog("%s no longer wants to recieve audio from %s\n",cltptr->clnt_config.name,ctmp->clnt_config.name);
                   ctmp->clnt_alist = unlink_slist(ctmp->clnt_alist,cltptr);
                }
	     }
	  }
       }
       ciptr++;
    }
}

void proc_sig()
{
    my_perror("Kill Signal received\n");
    exit(1);
}

void do_timer()
{
    static short refrefkcnt = 0;
    static short vatinterval = 0;
    client *ctmp,*ctmp1;
    maven_client *mtmp,*mtmp1;
    struct in_addr   in;

    /* delete any clients that have timed out */
       
    ctmp = chead;
    while (ctmp != NULL)
    {
       ctmp->clnt_stimer++;

       /* delete any regular or BCC_CLIENTS clients that have timed out */

       if ((ctmp->clnt_flags & (CLIENT | BCC_CLIENT)) != 0)
       {
          if (ctmp->clnt_rtimer++ > TIMEOUT)
          {
	     ctmp1 = ctmp;
             ctmp = ctmp->clnt_nptr;
	     if (ctmp1->clnt_flags & CLIENT)
	        dolog("client timeout -- deleting\n");
             else
	        dolog("BCC client timeout -- deleting\n");
	     delete_client(ctmp1);
          }
	  else
	     ctmp = ctmp->clnt_nptr;
       }
       else
          /* delete any clients that originated from another reflector   */
          if ((ctmp->clnt_flags & (BCC_ORIGIN | REF1_ORIGIN | REF2_ORIGIN | REF3_ORIGIN))  != 0)
          {
             if (ctmp->clnt_rtimer++ > (2*TIMEOUT))
             {
	        ctmp1 = ctmp;
                ctmp = ctmp->clnt_nptr;
	        if (ctmp1->clnt_flags & BCC_ORIGIN)
	           dolog("BCC client's client timeout -- deleting\n");
                else
	           dolog("REF client's client timeout -- deleting\n");
	        delete_client(ctmp1);
             }
	     else
                ctmp = ctmp->clnt_nptr;
          }
	  else
             ctmp = ctmp->clnt_nptr;
    }

    /* send keep alive packets to any BCC servers that we have */

    if (++refrefkcnt > KALIVE)
    {
       refrefkcnt = 0;
       open_bcc_servers(); 
       open_ucast_ref(); 
    }


    vatinterval++;
    if (vatinterval == 11)
       vatinterval = 0;

    mtmp = mhead;

    while (mtmp != NULL)
    {
      if (mtmp->mvn_rtimer++  > 50)
      {
	 in.s_addr =  mtmp->mvn_addr.addr;
	 dolog("Deleting Maven client at %s\n", inet_ntoa(in));

	 if (mtmp == mhead)
	 {
	    mhead = mtmp->mvn_nptr;
	    mtmp1 = mtmp;
            mtmp = mtmp->mvn_nptr;
            free(mtmp1);
         }
         else
	 {
	    mtmp1 = mhead;
	    while (mtmp1->mvn_nptr != mtmp)
	       mtmp1 = mtmp1->mvn_nptr;
            mtmp1->mvn_nptr = mtmp->mvn_nptr;
	    mtmp1 = mtmp;
            mtmp = mtmp->mvn_nptr;
            free(mtmp1);
         }
      }
      else
      {
	 if (vatinterval == 10)
	    vat_id(mtmp);

         mtmp = mtmp->mvn_nptr;
      }
    }

}

vat_id(mptr)
    maven_client *mptr;
{
    static char         buf[MAXMSG];
    struct CtrlMsgHdr   *vatIdmsg;
    int			msglen;
    struct in_addr      in;
    struct sockaddr_in  mvnskt;

#ifdef DEBUG
    if (debug)
    {
       in.s_addr = mptr->mvn_addr.addr;
       printf("sending id message to maven at %s\n",inet_ntoa(in));
    }
#endif

    vatIdmsg = (struct CtrlMsgHdr *) buf;
    vatIdmsg->flags = 0;
    vatIdmsg->type = CTRL_TYPE_ID;
    vatIdmsg->confid = 0;
    strcpy(vatIdmsg->idmsg,"CU-Reflector");
    vatIdmsg->idmsg[strlen("CU-Reflector")] = 0;
    msglen = CTRLMSGSIZE + strlen("CU-Reflector") + 1;

    mvnskt.sin_family = AF_INET;
    mvnskt.sin_port = mptr->mvn_addr.port + 1;
    mvnskt.sin_addr.s_addr = mptr->mvn_addr.addr;

    if (sendto(maven_cntl_sock,buf,msglen,0,&mvnskt,sizeof(struct sockaddr_in)) != msglen)
       dolog("Maven id sendto error\n");

}

void my_perror(mesg)
   char *mesg;
{
    extern char *sys_errlist[];
    extern int errno;

    dolog("%s : %s: FATAL ERROR: EXITING", mesg, sys_errlist[errno]);
    printf("%s : %s: FATAL ERROR: EXITING", mesg, sys_errlist[errno]);
    exit(1);
}

dolog(va_alist)
va_dcl
{
    static int line_cnt = 0;
    char    *fmt,*tmp;
    va_list args;
    struct timeval tp;
    struct timezone tzp;

    if (log_limit == 0)
       return;

    if (log_limit != -1)
    {
       if (line_cnt++ > log_limit)
       {
	  line_cnt = 0;
          fclose(log_file);
          if ((log_file = fopen(tracefile,"w")) != NULL)
          {
             setlinebuf(log_file);
             dolog("open_log file: %s\n",tracefile);
          }
          else
             log_limit = 0;
       }
    }


    gettimeofday(&tp, &tzp);
    tmp = ctime(&tp.tv_sec);
    tmp[24] = ' ';

    va_start(args);

    fmt = va_arg(args, char *);

    if (!log_file)
    {
        return;
    }

    if (fmt && *fmt)
    {
        (void) fprintf(log_file, "%s ", tmp);
        (void) vfprintf(log_file, fmt, args);
#ifdef DEBUG
       if (debug)
          (void) vfprintf(stdout, fmt, args);
#endif
    }
    va_end(args);
}

slist *unlink_slist(shead,cltptr)
    slist *shead;
    client *cltptr;
{
    slist *stmp;

    if (shead == NULL)
       return(NULL);
    else
       if (shead->snd_client == cltptr)
       {
          stmp = shead->snd_nptr;
          free_slist(shead);
          return(stmp);
       }

    shead->snd_nptr = unlink_slist(shead->snd_nptr,cltptr);
    return(shead);
}



client *unlink_client(chead,cltptr)
    client *chead,*cltptr;
{
    if (chead == NULL)
       return(NULL);
    else
       if (chead == cltptr)
       {
          cltptr->clnt_flags |= DELETE;
          return(chead->clnt_nptr);
       }

    chead->clnt_nptr = unlink_client(chead->clnt_nptr,cltptr);
    return(chead);
}


get_my_addr(myaddr)
    struct  sockaddr *myaddr;
{
    int n_interfaces, i;
    char *buf;
    struct ifconf ifc;
    struct ifreq *ifr;
    int s_in;

    bzero((caddr_t) myaddr,socksize(&ifr->ifr_addr));

    if ((s_in = socket(PF_INET, SOCK_DGRAM, 0)) < 0)
    {
        my_perror("if_getlist: socket");
        exit(errno);
    }

    ifc.ifc_len = (MAXIF * sizeof (struct ifreq)) + sizeof(int);

    if ( !(buf = malloc(ifc.ifc_len)))
    {
        my_perror("if_getlist: malloc");
        exit(errno);
    }

    ifc.ifc_buf = buf;

    if (ioctl(s_in, SIOCGIFCONF, (caddr_t) &ifc) < 0)
    {
        my_perror("if_getlist: ioctl(SIOCGIFCONF)");
        exit(errno);
    }

    n_interfaces = ifc.ifc_len / sizeof (struct ifreq);

    ifr = (struct ifreq *) ifc.ifc_buf;
    for (i = 0; i < n_interfaces; i++)
    {
        if (ifr->ifr_addr.sa_family == AF_INET)
        {
           bcopy((caddr_t) &ifr->ifr_addr, (caddr_t) myaddr, socksize(&ifr->ifr_addr));
           break;
        }
        ifr = (struct ifreq *) ((caddr_t) ifr + (sizeof(*ifr) + socksize(&ifr->ifr_addr) - sizeof (ifr->ifr_addr)));
    }
    (void) free(buf);
}


short get_type(csock)
    struct sockaddr_in  csock;
{
    short cnt;

    for (cnt=0; cnt<restrict_cnt; cnt++)
       if (restrict_list[cnt] == csock.sin_addr.s_addr)
          return(CLIENT);

    for (cnt=0; cnt<bcc_srv_cnt; cnt++)
       if (bcc_server_list[cnt] == csock.sin_addr.s_addr)
          return(BCC_SERVER);

    for (cnt=0; cnt<bcc_clt_cnt; cnt++)
       if (bcc_client_list[cnt] == csock.sin_addr.s_addr)
          return(BCC_CLIENT);

    for (cnt=0; cnt<ref_ucast_cnt; cnt++)
       if (ref_ucast_list[cnt] == csock.sin_addr.s_addr)
          return(REF3_SERVER);

    if (restrict_cnt != 0)
       return(-1);
    
    return(CLIENT);
}


distribute(vidptr,cltptr,all)
   VideoPacketHeader *vidptr;
   client *cltptr;
   char   all;
{
    client         *ctmp;
    unsigned char  *cptr;
    maven_client   *mtmp;
    slist          *sptr;
    struct in_addr in;
    int            len,cnt;
    short          ref_ucast_sent[MAXSRCREF];    


    if (all == FALSE)
    {
       for (cnt=0; cnt < MAXSRCREF; cnt++)
          ref_ucast_sent[cnt] = 0;

       sptr = cltptr->clnt_vlist;
       while (sptr)
       {
	  if (sptr->snd_client->clnt_flags & REF3_ORIGIN)
	  {
             if ((ctmp = sptr->snd_client->clnt_pptr) == NULL)
             {
                dolog("REF3_ORIGIN client without a parent in distribute\n");
		delete_client(cltptr);
		return;
             }

	     if ((ctmp->clnt_flags & REF3_SERVER) == 0)
             {
                dolog("REF3_ORIGIN client's parent is not a REF3_SERVER\n");
		delete_client(cltptr);
		return;
             }

             if ((ref_ucast_sent[ctmp->clnt_id] == 0) && (cltptr->clnt_flags & REF3_ORIGIN) == 0)
	     {
	        ref_ucast_sent[ctmp->clnt_id] = 1;
                write_pkt(vidptr,ctmp);
             }

             sptr = sptr->snd_nptr;
	     continue;
	  }

          write_pkt(vidptr,sptr->snd_client);
          sptr = sptr->snd_nptr;
       }

       ctmp = chead;
       while (ctmp != NULL)
       {
          if (ctmp->clnt_flags & BCC_CLIENT)
             write_pkt(vidptr,ctmp);
          ctmp = ctmp->clnt_nptr;
       }
    }
    else
    {
       ctmp = chead;
       while (ctmp != NULL)
       {
	  if ((ctmp->clnt_flags & CLIENT) && ((self_reflect == 1) || (cltptr != ctmp)))
             write_pkt(vidptr,ctmp);
          else
	     if ((ctmp->clnt_flags & REF3_SERVER) && ((cltptr->clnt_flags & (REF3_SERVER | REF3_ORIGIN)) == 0))
                write_pkt(vidptr,ctmp);
	     else
	        if ((ctmp->clnt_flags & BCC_CLIENT) && (cltptr != ctmp))
                   write_pkt(vidptr,ctmp);
          ctmp = ctmp->clnt_nptr;
       }
    }

#ifdef MULTI
    if (rfout_mcast.sin_addr.s_addr)
       rf_mcast_wrt(vidptr);

    if ((cltptr->clnt_flags & (REF2_SERVER | REF2_ORIGIN)) == 0)
       if (inout_mcast.sin_addr.s_addr)
          inout_mcast_wrt(vidptr);

    if (nv_mcast.sin_addr.s_addr)
       nv_mcast_wrt(vidptr,cltptr);
#endif
}

distribute_audio(vidptr,cltptr)
   VideoPacketHeader *vidptr;
   client *cltptr;
{
    client         *ctmp;
    unsigned char  *cptr;
    maven_client   *mtmp;
    slist          *sptr;
    struct in_addr in;
    int            len,cnt;
    short          ref_ucast_sent[MAXSRCREF];    

    if (cltptr->clnt_talker == 0)
       dolog("%s is speaking\n",cltptr->clnt_config.name);

    if (cltptr->clnt_talker++ > 50)
       cltptr->clnt_talker = 0;

    if (cltptr->clnt_config.sendMode == 0)
    {
       ctmp = chead;
       while (ctmp != NULL)
       {
          if (ctmp->clnt_flags & CLIENT)
             if ((ctmp->clnt_config.flags & REC_AUDIO) && (ctmp->clnt_config.flags & WANT_LURCKERS))
                write_pkt(vidptr,ctmp);
         ctmp = ctmp->clnt_nptr;
       }
    }
    else
    {
       for (cnt=0; cnt < MAXSRCREF; cnt++)
          ref_ucast_sent[cnt] = 0;

       sptr = cltptr->clnt_alist;
       while (sptr)
       {
	  if (sptr->snd_client->clnt_flags & REF3_ORIGIN)
	  {
             if ((ctmp = sptr->snd_client->clnt_pptr) == NULL)
             {
                dolog("REF3_ORIGIN client without a parent in distribute_audio\n");
		delete_client(cltptr);
		return;
             }

	     if ((ctmp->clnt_flags & REF3_SERVER) == 0)
             {
                dolog("REF3_ORIGIN client's parent is not a REF3_SERVER\n");
		delete_client(cltptr);
		return;
             }

             if ((ref_ucast_sent[ctmp->clnt_id] == 0) && (cltptr->clnt_flags & REF3_ORIGIN) == 0)
	     {
	        ref_ucast_sent[ctmp->clnt_id] = 1;
                write_pkt(vidptr,ctmp);
             }

             sptr = sptr->snd_nptr;
	     continue;
	  }

          write_pkt(vidptr,sptr->snd_client);
          sptr = sptr->snd_nptr;
       }
    }

    ctmp = chead;
    while (ctmp != NULL)
    {
       if (ctmp->clnt_flags & BCC_CLIENT)
          write_pkt(vidptr,ctmp);
       ctmp = ctmp->clnt_nptr;
    }

#ifdef MULTI
    if (rfout_mcast.sin_addr.s_addr)
       rf_mcast_wrt(vidptr);

    if ((cltptr->clnt_flags & (REF2_SERVER | REF2_ORIGIN)) == 0)
       if (inout_mcast.sin_addr.s_addr)
          inout_mcast_wrt(vidptr);
#endif

    cptr = ((unsigned char *) vidptr) + HEADERLEN;
    len = vidptr->len - HEADERLEN;
    mtmp = mhead;

    while (mtmp != NULL)
    {
       if (sendto(maven_sock,cptr,len,0,&mtmp->mvn_addr,sizeof(struct sockaddr_in)) != len)
          dolog("maven sendto error in distribute\n");
       mtmp = mtmp->mvn_nptr;
    }
}

short deny(csock)
    struct sockaddr_in  csock;
{
    short cnt;

    for (cnt=0; cnt<deny_cnt; cnt++)
       if (deny_list[cnt] == csock.sin_addr.s_addr)
          return(1);

   return(0);
}
